"""
Default ConfigUI for W8-E2000D40 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dZHkAoSDxQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABUCZUAWVQ3'
    b'VzdG9tIFJlY29yZGluZyBUYWdzUAGDF/9/UQSAUSSBQFAV8UtleXdvcmRzIGZvciBvcmdh'
    b'bml6aW5nIGFuZCBpZGVudGlmeWluZyByZWNvcmRpbmdzLCBzZXBhcmF0ZWQgYnkgY29tbW'
    b'Egb3Igc3BhY2UuIExpbWl0IGlzIDIwIHRhZ3MsIDY0IGNoYXJhY3RlcnMuQAX7UAWTUmVj'
    b'b3JkaW5nIERpcmVjdG9yeVABgxT/f1EEhlJFQ09SRFEkgRBQFJNDb25maWdbMHgxNkZGN0'
    b'ZdPT0wUBW5RGlyZWN0b3J5IGZvciBzZXJpYWxpemVkIHJlY29yZGluZ3MuIDE2IGNoYXJh'
    b'Y3RlciBtYXhpbXVtQAVAo1AFlVJlY29yZGluZyBGaWxlIFByZWZpeFABgxX/f1EkgRBQFe'
    b'hQcmVmaXggZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5nIG5hbWVzLiBJZiBibGFuaywgZGV2'
    b'aWNlIHNlcmlhbCBudW1iZXIgd2lsbCBiZSB1c2VkLiAxNiBjaGFyYWN0ZXIgbWF4aW11bV'
    b'AUk0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGB'
    b'AVAVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIH'
    b'RoZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1p'
    b'ZG5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZW'
    b'N0b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51'
    b'bWJlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBY'
    b'pUaW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQKlQBYtCdXR0b24gTW9kZVABgxD/f1EB'
    b'gQBQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3'
    b'IgYmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBT'
    b'ZWNvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BBB41QBYpTdGFydC1Pbmx5QBdBFFAFjlBsdW'
    b'ctSW4gQWN0aW9uUAGDCv9/UQGBAVAVrVNldCBhY3Rpb24gZm9yIGRldmljZSB3aGVuIHBs'
    b'dWdnZWQgaW4gdG8gVVNCLkEHtVAFskltbWVkaWF0ZWx5IHN0b3AgcmVjb3JkaW5nIGFuZC'
    b'BhcHBlYXIgYXMgVVNCIGRyaXZlQQezUAWwQ29tcGxldGUgcmVjb3JkaW5nIGJlZm9yZSBh'
    b'cHBlYXJpbmcgYXMgVVNCIGRyaXZlQQewUAWtSWdub3JlOiBzdG9wIHJlY29yZGluZyB3aG'
    b'VuIGJ1dHRvbiBpcyBwcmVzc2VkQQelUAWiU3RhcnQgcmVjb3JkaW5nIHdoZW4gVVNCIGNv'
    b'bm5lY3RlZEAj51ABgwv/f1AV3kVudGVyIGxvY2FsIHRpbWV6b25lJ3Mgb2Zmc2V0IGZyb2'
    b'0gVVRDIHRpbWUuIEltcGFjdHMgdGltZXN0YW1wIGluZm9ybWF0aW9uIG9uIHJlY29yZGlu'
    b'ZyBmaWxlcy5AL4BAGEIIUAWMRW5hYmxlIFdpLUZpUAGDGP9/UQGBAUAn7FABgxn/f1EBgQ'
    b'FBB6lQBaZVcGxvYWQgcmVjb3JkaW5nIHRvIGNsb3VkIGFmdGVyIGZpbmlzaFAVs1NlbGVj'
    b'dCB0byB1cGxvYWQgZWFjaCBmaWxlIHdoZW4gcmVjb3JkaW5nIGNvbXBsZXRlc0AnQLZQAY'
    b'Ma/39RAYEAQQejUAWgRGVsZXRlIGZyb20gc3RvcmFnZSBhZnRlciB1cGxvYWRQFUCCUmVj'
    b'b3JkaW5nIGZpbGUgd2lsbCBiZSBkZWxldGVkIGFmdGVyIGEgc3VjY2Vzc2Z1bCB1cGxvYW'
    b'QsIGFuZCBtb3ZlZCB0byBhIHN1YmZvbGRlciBsYWJlbGVkICdmYWlsZWQnIGluIHRoZSBj'
    b'YXNlIG9mIGEgZmFpbGVkIHVwbG9hZEAnQMJQAYMd/39RAYEAQQepUAWmS2VlcCB0aGUgV2'
    b'ktRmkgbW9kdWxlIG9uIGFuZCBjb25uZWN0ZWRQFUCISWYgY2hlY2tlZCwgdGhlIFdpLUZp'
    b'IG1vZHVsZSB3aWxsIHN0YXkgb24gYW5kIGNvbm5lY3RlZC4gIEFsbG93cyBmb3IgZmFzdG'
    b'VyIHVwbG9hZCB0aW1lcyB3aXRoIHRoZSBleHBlbnNlIG9mIGluY3JlYXNlZCBwb3dlciBj'
    b'b25zdW1wdGlvbkAPgEBPgEAoSOlQBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQAY'
    b'MS/39RAYEAUBVAj0RlbGF5IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZSwg'
    b'dGhlbiB3YWl0IGZvciB0cmlnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0ZX'
    b'Igd2FpdCBvciBvbiB0cmlnZ2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQeV'
    b'UAWSRGVsYXkgVGhlbiBUcmlnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0YX'
    b'J0IGF0IFRpbWVQAYMP/39QFdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdWF0'
    b'aW5nIG90aGVyIHRyaWdnZXIgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5nIG'
    b'RlbGF5QBGuUAWUUmVjb3JkaW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb25k'
    b'c0ARQJBQBZhSZWNvcmRpbmcgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25kc1'
    b'EhgwFRgFAV2FRpbWUgYmVmb3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZSBi'
    b'ZXR3ZWVuIHJlY29yZGluZ3MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlSZW'
    b'NvcmRpbmcgRmlsZSBTaXplIExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZXRy'
    b'aWdnZXJQAYMO/39QFLBDb25maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDExRk'
    b'Y3Rl09PW51bGxQFc9XaGVuIHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlciBy'
    b'ZWNvcmRpbmcgdGltZSBvciBzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQgZm'
    b'9yIEFsbCBTZW5zb3IgQ29uZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbmcg'
    b'd2lsbCBvbmx5IHN0YXJ0IHdoZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lzZS'
    b'ByZWNvcmRpbmcgd2lsbCBzdGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG1l'
    b'dEAYQO5QBZBQcmVzc3VyZSBUcmlnZ2VyUAGDBQAkUQCBAECC5VAFlVByZXNzdXJlIFRyaW'
    b'dnZXIsIExvd1ABgwMAJFEBgwFfkFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbXBs'
    b'aW5nIHdoZW4gb3V0c2lkZSBwcmVzc3VyZSB3aW5kb3cuQILmUAWWUHJlc3N1cmUgVHJpZ2'
    b'dlciwgSGlnaFABgwQAJFEBgwGtsFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbXBs'
    b'aW5nIHdoZW4gb3V0c2lkZSBwcmVzc3VyZSB3aW5kb3cuQBhA+1AFk1RlbXBlcmF0dXJlIF'
    b'RyaWdnZXJQAYMFASRRAIEAQGPqUAWYVGVtcGVyYXR1cmUgVHJpZ2dlciwgTG93UAGDAwEk'
    b'UQKB9lAlgUNREoHYUSKBUFAVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZG'
    b'UgdGVtcGVyYXR1cmUgd2luZG93LkBj61AFmVRlbXBlcmF0dXJlIFRyaWdnZXIsIEhpZ2hQ'
    b'AYMEASRRAoEjUCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3'
    b'V0c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQBhBqVAFnEFjY2VsZXJhdGlvbiBUcmlnZ2Vy'
    b'ICgyMDAwZylQAYMF/whQEYMH/yBRAIEAUCSAUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAwQE'
    b'NAq1AFmUFjY2VsZXJhdGlvbiBUcmlnZ2VyLCBMb3dQAYMD/whRAoL5mlETiMCfQAAAAAAA'
    b'USOIQJ9AAAAAAABQFeBIaWdoIHBvd2VyIHRyaWdnZXIgd2l0aCBwcmUtdHJpZ2dlciBkYX'
    b'RhLiBOb3QgcmVjb21tZW5kZWQgZm9yIGxvbmcgd2FpdHMgd2l0aG91dCBleHRlcm5hbCBw'
    b'b3dlci5SA4g/r0AfQB9ILkBDQKxQBZpBY2NlbGVyYXRpb24gVHJpZ2dlciwgSGlnaFABgw'
    b'T/CFECggZmUROIwJ9AAAAAAABRI4hAn0AAAAAAAFAV4EhpZ2ggcG93ZXIgdHJpZ2dlciB3'
    b'aXRoIHByZS10cmlnZ2VyIGRhdGEuIE5vdCByZWNvbW1lbmRlZCBmb3IgbG9uZyB3YWl0cy'
    b'B3aXRob3V0IGV4dGVybmFsIHBvd2VyLlIDiD+vQB9AH0guQBhBo1AFmkFjY2VsZXJhdGlv'
    b'biBUcmlnZ2VyICg0MGcpUAGDB/9QUBXUTG93IHBvd2VyIHRyaWdnZXIgdXNpbmcgZGlnaX'
    b'RhbCBhY2NlbGVyb21ldGVyIHdpdGggREMgcmVzcG9uc2UuIE5vIHByZS10cmlnZ2VyIGRh'
    b'dGEuUQCBAFARgwX/CFARgwf/IFAkgFAUlENvbmZpZ1sweDFmZjUwXSA9PSAwQENAslAFlk'
    b'FjY2VsZXJhdGlvbiBUaHJlc2hvbGRQAYME/1BRAYIgAFIDiD9EABQAFAAUUSOIQEQAAAAA'
    b'AABRE4BQFfJMb3cgcG93ZXIgdHJpZ2dlci4gVGhlIG1pbmltdW0gYWNjZWxlcmF0aW9uIC'
    b'hwb3NpdGl2ZSBvciBuZWdhdGl2ZSkgdG8gdHJpZ2dlciBhIHJlY29yZGluZy4gTXVzdCBi'
    b'ZSBncmVhdGVyIHRoYW4gMC5AJ8ZQAYMF/1BRAYEHQQeRUAWOWCBBeGlzIFRyaWdnZXJBB5'
    b'FQBY5ZIEF4aXMgVHJpZ2dlckEHkVAFjlogQXhpcyBUcmlnZ2VyQA+AQE+AQChH5FAFjE1l'
    b'YXN1cmVtZW50c0AIQlxQBZkyMDAwZyBBY2NlbGVyYXRpb24gKENoIDgpUCSAQCdAm1EBgQ'
    b'9QAYMB/whQFaBFbmFibGUvRGlzYWJsZSBhY2NlbGVyYXRpb24gYXhlc0EHmFAFkUVuYWJs'
    b'ZSBYIChDaCA4LjApUQGBAEEHmFAFkUVuYWJsZSBZIChDaCA4LjEpUQGBAUEHmFAFkUVuYW'
    b'JsZSBaIChDaCA4LjIpUQGBAkEHmlAFk0VuYWJsZSBNaWMgKENoIDguMylRAYEDQBFAmlAF'
    b'i1NhbXBsZSBSYXRlUAGDAv8IUBXYU2V0IG1haW4gYWNjZWxlcmF0aW9uIHNhbXBsZSByYX'
    b'RlIGZyb20gMTAgdG8gMjAwMDAgSHouCkhpZ2hlciByYXRlIHJlZHVjZXMgYmF0dGVyeSBs'
    b'aWZlLlAUlUNvbmZpZ1sweDAxRkYwOF0gPT0gMFEBghOIUCWCSHpREYEKUSGCTiBAEUD8UA'
    b'WjT3ZlcnJpZGUgQW50aWFsaWFzaW5nIEZpbHRlciBDdXRvZmZQAYMI/whRAYID6FAlgkh6'
    b'USGCYahQFUClRGVmYXVsdCwgd2hlbiBub3QgY2hlY2tlZCwgb2Ygb25lLWZpZnRoIHNhbX'
    b'BsZSBmcmVxdWVuY3kuIEN1dC1vZmYgZnJlcXVlbmN5IGhhcyAyOSUgYXR0ZW51YXRpb24u'
    b'CjV0aCBvcmRlciBCdXR0ZXJ3b3J0aC4gU29tZSBhdHRlbnVhdGlvbiBhdCA2MCUgb2YgY3'
    b'V0LW9mZiBmcmVxdWVuY3kuUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAwQAhBHFAFmzQwZyBE'
    b'QyBBY2NlbGVyYXRpb24gKENoIDgwKUAnn1ABgwH/UFEBgQFBB5JQBY9FbmFibGUgQWxsIE'
    b'F4ZXNAF0DYUAWLU2FtcGxlIFJhdGVQAYMC/1BRAYIB9FAlgkh6UBXeTG93IHBhc3MgZmls'
    b'dGVyIGlzIHNldCB0byAyNSUgb2YgZGF0YSByYXRlLCBzbyBhIGRhdGEgcmF0ZSBvZiAyMD'
    b'AwSHogaGFzIGEgTFAgZmlsdGVyIGF0IDUwMCBIelAUk25vdCBDb25maWdbMHgxZmY1MF1B'
    b'B4VRAYIPoEEHhVEBggfQQQeFUQGCA+hBB4VRAYIB9EEHhFEBgfpBB4RRAYF9QQeEUQGBP0'
    b'EHhFEBgSBBB4RRAYEQQAhBMFAFqENvbnRyb2wgUGFkIFRlbXBlcmF0dXJlL1ByZXNzdXJl'
    b'IChDaCA1OSlAJ0CbUQGBA1ABgwH/O1AVpUVuYWJsZS9EaXNhYmxlIHNlbnNvcnMgb24gY2'
    b'9udHJvbCBwYWRBB5xQBZlFbmFibGUgUHJlc3N1cmUgKENoIDU5LjApQQefUAWcRW5hYmxl'
    b'IFRlbXBlcmF0dXJlIChDaCA1OS4xKUEHpVAFokVuYWJsZSBSZWxhdGl2ZSBIdW1pZGl0eS'
    b'AoQ2ggNTkuMilAEeNQBYtTYW1wbGUgUmF0ZVABgwL/O1AUlENvbmZpZ1sweDFmZjNCXSA9'
    b'PSAwUBWkU2V0IHNhbXBsZSBmcmVxdWVuY3kgZnJvbSAxIHRvIDEwIEh6UQGBClAlgkh6UR'
    b'GBAVEhgQpACEHNUAWZSW5lcnRpYWwgTWVhc3VyZW1lbnQgVW5pdEAHQJZQBZBBY3F1aXNp'
    b'dGlvbiBNb2RlUAGDAf8rUCWBIFEBgRBBB4pQBYNPZmZRAYEAQQeoUAWhQWJzb2x1dGUgT3'
    b'JpZW50YXRpb24gKFF1YXRlcm5pb24pUQGBCEEHqFAFoVJlbGF0aXZlIE9yaWVudGF0aW9u'
    b'IChRdWF0ZXJuaW9uKVEBgRBBB49QBYhSb3RhdGlvblEBgQJAAUCGUAWVT3JpZW50YXRpb2'
    b'4gRGF0YSBSYXRlUAGDAggrUQGBZFAlgkh6URGBAVEhgchQFb1TYW1wbGUgcmF0ZSBmb3Ig'
    b'ZGlyZWN0aW9uYWwgZGF0YSBmcm9tIDEgdG8gMjAwIEh6LiBDaGFubmVsIDY1UBSUQ29uZm'
    b'lnWzB4MDFGRjJCXSA8IDhAAUCJUAWSUm90YXRpb24gRGF0YSBSYXRlUAGDAgIrUQGBZFAl'
    b'gkh6URGBAVEhgchQFb5TZXQgc2FtcGxlIHJhdGUgZm9yIHJvdGF0aW9uIGRhdGEgZnJvbS'
    b'AxIHRvIDIwMCBIei4gQ2hhbm5lbCA0N1AUmUNvbmZpZ1sweDAxRkYyQl0gJiAyID09IDBA'
    b'CECQUAWgQ29udHJvbCBQYWQgTGlnaHQgU2Vuc29yIChDaCA3NilAJ+pQAYMB/0xQFb5Fbm'
    b'FibGUvRGlzYWJsZSBsaWdodCBzZW5zb3IgYXQgNCBIeiAtIE5PVEU6IERJU0FCTEVEIEZP'
    b'UiBBTFBIQUEHoFAFnUVuYWJsZSBMaWdodCBTZW5zb3IgKENoIDc2LjApQAhAslAFk0JNRz'
    b'I1MCBHeXJvIChDaCA4NClAJ59QAYMB/1RRAYEAQQeSUAWPRW5hYmxlIEFsbCBBeGVzQBf3'
    b'UAWLU2FtcGxlIFJhdGVQAYMC/1RRAYIBkFAlgkh6UBSTbm90IENvbmZpZ1sweDFmZjU0XU'
    b'EHhVEBggyAQQeFUQGCBkBBB4VRAYIDIEEHhVEBggGQQQeEUQGByEEHhFEBgWRBB4RRAYEy'
    b'QQeEUQGBGUEHhFEBgQxAD4BAT4BKCIBKKIBKSIA='
)



"""
Default ConfigUI for LOG-0004-02kG-DC data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dVDkAoRYFQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABfpQBZNSZW'
    b'NvcmRpbmcgRGlyZWN0b3J5UAGDFP9/UQSGUkVDT1JEUSSBCFAUk0NvbmZpZ1sweDE2RkY3'
    b'Rl09PTBQFbhEaXJlY3RvcnkgZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5ncy4gOCBjaGFyYW'
    b'N0ZXIgbWF4aW11bUAFQKJQBZVSZWNvcmRpbmcgRmlsZSBQcmVmaXhQAYMV/39RJIEIUBXn'
    b'UHJlZml4IGZvciBzZXJpYWxpemVkIHJlY29yZGluZyBuYW1lcy4gSWYgYmxhbmssIGRldm'
    b'ljZSBzZXJpYWwgbnVtYmVyIHdpbGwgYmUgdXNlZC4gOCBjaGFyYWN0ZXIgbWF4aW11bVAU'
    b'k0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGBAF'
    b'AVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIHRo'
    b'ZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1pZG'
    b'5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZWN0'
    b'b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51bW'
    b'JlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBYpU'
    b'aW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQJlQBYtCdXR0b24gTW9kZVABgxD/f1EBgQ'
    b'BQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3Ig'
    b'YmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBTZW'
    b'NvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BAF0EUUAWOUGx1Zy1JbiBBY3Rpb25QAYMK/39R'
    b'AYEBUBWtU2V0IGFjdGlvbiBmb3IgZGV2aWNlIHdoZW4gcGx1Z2dlZCBpbiB0byBVU0IuQQ'
    b'e1UAWySW1tZWRpYXRlbHkgc3RvcCByZWNvcmRpbmcgYW5kIGFwcGVhciBhcyBVU0IgZHJp'
    b'dmVBB7NQBbBDb21wbGV0ZSByZWNvcmRpbmcgYmVmb3JlIGFwcGVhcmluZyBhcyBVU0IgZH'
    b'JpdmVBB7BQBa1JZ25vcmU6IHN0b3AgcmVjb3JkaW5nIHdoZW4gYnV0dG9uIGlzIHByZXNz'
    b'ZWRBB6VQBaJTdGFydCByZWNvcmRpbmcgd2hlbiBVU0IgY29ubmVjdGVkQCPnUAGDC/9/UB'
    b'XeRW50ZXIgbG9jYWwgdGltZXpvbmUncyBvZmZzZXQgZnJvbSBVVEMgdGltZS4gSW1wYWN0'
    b'cyB0aW1lc3RhbXAgaW5mb3JtYXRpb24gb24gcmVjb3JkaW5nIGZpbGVzLkAvgEAPgEBPgE'
    b'AoSOJQBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQAYMS/39RAYEAUBVAj0RlbGF5'
    b'IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZSwgdGhlbiB3YWl0IGZvciB0cm'
    b'lnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0ZXIgd2FpdCBvciBvbiB0cmln'
    b'Z2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQeVUAWSRGVsYXkgVGhlbiBUcm'
    b'lnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0YXJ0IGF0IFRpbWVQAYMP/39Q'
    b'FdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdWF0aW5nIG90aGVyIHRyaWdnZX'
    b'IgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5nIGRlbGF5QBGuUAWUUmVjb3Jk'
    b'aW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb25kc0ARQJBQBZhSZWNvcmRpbm'
    b'cgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25kc1EhgwFRgFAV2FRpbWUgYmVm'
    b'b3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZSBiZXR3ZWVuIHJlY29yZGluZ3'
    b'MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlSZWNvcmRpbmcgRmlsZSBTaXpl'
    b'IExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZXRyaWdnZXJQAYMO/39QFLBDb2'
    b'5maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDExRkY3Rl09PW51bGxQFc9XaGVu'
    b'IHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlciByZWNvcmRpbmcgdGltZSBvci'
    b'BzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQgZm9yIEFsbCBTZW5zb3IgQ29u'
    b'ZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbmcgd2lsbCBvbmx5IHN0YXJ0IH'
    b'doZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lzZSByZWNvcmRpbmcgd2lsbCBz'
    b'dGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG1ldEAYQO5QBZBQcmVzc3VyZS'
    b'BUcmlnZ2VyUAGDBQAkUQCBAECC5VAFlVByZXNzdXJlIFRyaWdnZXIsIExvd1ABgwMAJFEB'
    b'gwFfkFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZS'
    b'BwcmVzc3VyZSB3aW5kb3cuQILmUAWWUHJlc3N1cmUgVHJpZ2dlciwgSGlnaFABgwQAJFEB'
    b'gwGtsFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZS'
    b'BwcmVzc3VyZSB3aW5kb3cuQBhA+1AFk1RlbXBlcmF0dXJlIFRyaWdnZXJQAYMFASRRAIEA'
    b'QGPqUAWYVGVtcGVyYXR1cmUgVHJpZ2dlciwgTG93UAGDAwEkUQKB9lAlgUNREoHYUSKBUF'
    b'AVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgdGVtcGVyYXR1cmUgd2lu'
    b'ZG93LkBj61AFmVRlbXBlcmF0dXJlIFRyaWdnZXIsIEhpZ2hQAYMEASRRAoEjUCWBQ1ESgd'
    b'hRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZSB0ZW1wZXJhdHVy'
    b'ZSB3aW5kb3cuQBhBqVAFnEFjY2VsZXJhdGlvbiBUcmlnZ2VyICgyMDAwZylQAYMF/whQEY'
    b'MH/yBRAIEAUCSAUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAwQENAq1AFmUFjY2VsZXJhdGlv'
    b'biBUcmlnZ2VyLCBMb3dQAYMD/whRAoL5mlETiMCfQAAAAAAAUSOIQJ9AAAAAAABQFeBIaW'
    b'doIHBvd2VyIHRyaWdnZXIgd2l0aCBwcmUtdHJpZ2dlciBkYXRhLiBOb3QgcmVjb21tZW5k'
    b'ZWQgZm9yIGxvbmcgd2FpdHMgd2l0aG91dCBleHRlcm5hbCBwb3dlci5SA4g/r0AfQB9ILk'
    b'BDQKxQBZpBY2NlbGVyYXRpb24gVHJpZ2dlciwgSGlnaFABgwT/CFECggZmUROIwJ9AAAAA'
    b'AABRI4hAn0AAAAAAAFAV4EhpZ2ggcG93ZXIgdHJpZ2dlciB3aXRoIHByZS10cmlnZ2VyIG'
    b'RhdGEuIE5vdCByZWNvbW1lbmRlZCBmb3IgbG9uZyB3YWl0cyB3aXRob3V0IGV4dGVybmFs'
    b'IHBvd2VyLlIDiD+vQB9AH0guQBhBnFAFmkFjY2VsZXJhdGlvbiBUcmlnZ2VyICgxNmcpUA'
    b'GDB/8gUBXUTG93IHBvd2VyIHRyaWdnZXIgdXNpbmcgZGlnaXRhbCBhY2NlbGVyb21ldGVy'
    b'IHdpdGggREMgcmVzcG9uc2UuIE5vIHByZS10cmlnZ2VyIGRhdGEuUQCBAFARgwX/CFAkgF'
    b'AUlENvbmZpZ1sweDFmZjIwXSA9PSAwQENAsVAFlkFjY2VsZXJhdGlvbiBUaHJlc2hvbGRQ'
    b'AYME/yBRAYFQUgOIP7AQEBAQEBBRI4hAMAAAAAAAAFETgFAV8kxvdyBwb3dlciB0cmlnZ2'
    b'VyLiBUaGUgbWluaW11bSBhY2NlbGVyYXRpb24gKHBvc2l0aXZlIG9yIG5lZ2F0aXZlKSB0'
    b'byB0cmlnZ2VyIGEgcmVjb3JkaW5nLiBNdXN0IGJlIGdyZWF0ZXIgdGhhbiAwLkAnxlABgw'
    b'X/IFEBgQdBB5FQBY5YIEF4aXMgVHJpZ2dlckEHkVAFjlkgQXhpcyBUcmlnZ2VyQQeRUAWO'
    b'WiBBeGlzIFRyaWdnZXJAD4BAT4BAKEaWUAWMTWVhc3VyZW1lbnRzQAhCQlAFmTIwMDBnIE'
    b'FjY2VsZXJhdGlvbiAoQ2ggOClQJIBAJ0CBUQGBB1ABgwH/CFAVoEVuYWJsZS9EaXNhYmxl'
    b'IGFjY2VsZXJhdGlvbiBheGVzQQeZUAWWRW5hYmxlIFggQXhpcyAoQ2ggOC4wKUEHmVAFlk'
    b'VuYWJsZSBZIEF4aXMgKENoIDguMSlBB5lQBZZFbmFibGUgWiBBeGlzIChDaCA4LjIpQBFA'
    b'mlAFi1NhbXBsZSBSYXRlUAGDAv8IUBXYU2V0IG1haW4gYWNjZWxlcmF0aW9uIHNhbXBsZS'
    b'ByYXRlIGZyb20gMTAgdG8gMjAwMDAgSHouCkhpZ2hlciByYXRlIHJlZHVjZXMgYmF0dGVy'
    b'eSBsaWZlLlAUlUNvbmZpZ1sweDAxRkYwOF0gPT0gMFEBghOIUCWCSHpREYEKUSGCTiBAEU'
    b'D8UAWjT3ZlcnJpZGUgQW50aWFsaWFzaW5nIEZpbHRlciBDdXRvZmZQAYMI/whRAYID6FAl'
    b'gkh6USGCE4hQFUClRGVmYXVsdCwgd2hlbiBub3QgY2hlY2tlZCwgb2Ygb25lLWZpZnRoIH'
    b'NhbXBsZSBmcmVxdWVuY3kuIEN1dC1vZmYgZnJlcXVlbmN5IGhhcyAyOSUgYXR0ZW51YXRp'
    b'b24uCjV0aCBvcmRlciBCdXR0ZXJ3b3J0aC4gU29tZSBhdHRlbnVhdGlvbiBhdCA2MCUgb2'
    b'YgY3V0LW9mZiBmcmVxdWVuY3kuUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAwQAhAulAFmzE2'
    b'ZyBEQyBBY2NlbGVyYXRpb24gKENoIDMyKUAnn1ABgwH/IFEBgQFBB5JQBY9FbmFibGUgQW'
    b'xsIEF4ZXNAF/dQBYtTYW1wbGUgUmF0ZVABgwL/IFEBggGQUCWCSHpQFJNub3QgQ29uZmln'
    b'WzB4MWZmMjBdQQeFUQGCDIBBB4VRAYIGQEEHhVEBggMgQQeFUQGCAZBBB4RRAYHIQQeEUQ'
    b'GBZEEHhFEBgTJBB4RRAYEZQQeEUQGBDEAIQTBQBahDb250cm9sIFBhZCBUZW1wZXJhdHVy'
    b'ZS9QcmVzc3VyZSAoQ2ggNTkpQCdAm1EBgQNQAYMB/ztQFaVFbmFibGUvRGlzYWJsZSBzZW'
    b'5zb3JzIG9uIGNvbnRyb2wgcGFkQQecUAWZRW5hYmxlIFByZXNzdXJlIChDaCA1OS4wKUEH'
    b'n1AFnEVuYWJsZSBUZW1wZXJhdHVyZSAoQ2ggNTkuMSlBB6VQBaJFbmFibGUgUmVsYXRpdm'
    b'UgSHVtaWRpdHkgKENoIDU5LjIpQBHjUAWLU2FtcGxlIFJhdGVQAYMC/ztQFJRDb25maWdb'
    b'MHgxZmYzQl0gPT0gMFAVpFNldCBzYW1wbGUgZnJlcXVlbmN5IGZyb20gMSB0byAxMCBIel'
    b'EBgQpQJYJIelERgQFRIYEKQAhByVAFmUluZXJ0aWFsIE1lYXN1cmVtZW50IFVuaXRAB0CS'
    b'UAWQQWNxdWlzaXRpb24gTW9kZVABgwH/K1EBgRBBB4pQBYNPZmZRAYEAQQeoUAWhQWJzb2'
    b'x1dGUgT3JpZW50YXRpb24gKFF1YXRlcm5pb24pUQGBCEEHqFAFoVJlbGF0aXZlIE9yaWVu'
    b'dGF0aW9uIChRdWF0ZXJuaW9uKVEBgRBBB49QBYhSb3RhdGlvblEBgQJAAUCGUAWVT3JpZW'
    b'50YXRpb24gRGF0YSBSYXRlUAGDAggrUQGBZFAlgkh6URGBAVEhgchQFb1TYW1wbGUgcmF0'
    b'ZSBmb3IgZGlyZWN0aW9uYWwgZGF0YSBmcm9tIDEgdG8gMjAwIEh6LiBDaGFubmVsIDY1UB'
    b'SUQ29uZmlnWzB4MDFGRjJCXSA8IDhAAUCJUAWSUm90YXRpb24gRGF0YSBSYXRlUAGDAgIr'
    b'UQGBZFAlgkh6URGBAVEhgchQFb5TZXQgc2FtcGxlIHJhdGUgZm9yIHJvdGF0aW9uIGRhdG'
    b'EgZnJvbSAxIHRvIDIwMCBIei4gQ2hhbm5lbCA0N1AUmUNvbmZpZ1sweDAxRkYyQl0gJiAy'
    b'ID09IDBACPlQBaBDb250cm9sIFBhZCBMaWdodCBTZW5zb3IgKENoIDc2KUAn01EBgQFQAY'
    b'MB/0xQFaNFbmFibGUvRGlzYWJsZSBsaWdodCBzZW5zb3IgYXQgNCBIekEHoFAFnUVuYWJs'
    b'ZSBMaWdodCBTZW5zb3IgKENoIDc2LjApQA+AQE+ASgiASiiASkiA'
)



"""
Type definitions, primarily for type hinting.
"""

from pathlib import Path
from typing import AnyStr, Optional, NamedTuple, Union


# ==============================================================================
# Data-containing classes
# ==============================================================================

class Drive(NamedTuple):
    """ Representation of basic drive/volume information. Used internally. """
    path: Union[AnyStr, Path]
    label: Optional[str]
    sn: Optional[str]
    label: Optional[str]
    fs: Optional[str]
    type: Optional[str]


    def __str__(self):
        return str(self.path)


# ==============================================================================
# Type hinting definitions
# ==============================================================================

Epoch = Union[float, int]
Filename = Union[AnyStr, Path, Drive]





"""
Default ConfigUI for W8-D40 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dYR0AoSDxQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABUCZUAWVQ3'
    b'VzdG9tIFJlY29yZGluZyBUYWdzUAGDF/9/UQSAUSSBQFAV8UtleXdvcmRzIGZvciBvcmdh'
    b'bml6aW5nIGFuZCBpZGVudGlmeWluZyByZWNvcmRpbmdzLCBzZXBhcmF0ZWQgYnkgY29tbW'
    b'Egb3Igc3BhY2UuIExpbWl0IGlzIDIwIHRhZ3MsIDY0IGNoYXJhY3RlcnMuQAX7UAWTUmVj'
    b'b3JkaW5nIERpcmVjdG9yeVABgxT/f1EEhlJFQ09SRFEkgRBQFJNDb25maWdbMHgxNkZGN0'
    b'ZdPT0wUBW5RGlyZWN0b3J5IGZvciBzZXJpYWxpemVkIHJlY29yZGluZ3MuIDE2IGNoYXJh'
    b'Y3RlciBtYXhpbXVtQAVAo1AFlVJlY29yZGluZyBGaWxlIFByZWZpeFABgxX/f1EkgRBQFe'
    b'hQcmVmaXggZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5nIG5hbWVzLiBJZiBibGFuaywgZGV2'
    b'aWNlIHNlcmlhbCBudW1iZXIgd2lsbCBiZSB1c2VkLiAxNiBjaGFyYWN0ZXIgbWF4aW11bV'
    b'AUk0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGB'
    b'AVAVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIH'
    b'RoZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1p'
    b'ZG5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZW'
    b'N0b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51'
    b'bWJlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBY'
    b'pUaW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQKlQBYtCdXR0b24gTW9kZVABgxD/f1EB'
    b'gQBQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3'
    b'IgYmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBT'
    b'ZWNvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BBB41QBYpTdGFydC1Pbmx5QBdBFFAFjlBsdW'
    b'ctSW4gQWN0aW9uUAGDCv9/UQGBAVAVrVNldCBhY3Rpb24gZm9yIGRldmljZSB3aGVuIHBs'
    b'dWdnZWQgaW4gdG8gVVNCLkEHtVAFskltbWVkaWF0ZWx5IHN0b3AgcmVjb3JkaW5nIGFuZC'
    b'BhcHBlYXIgYXMgVVNCIGRyaXZlQQezUAWwQ29tcGxldGUgcmVjb3JkaW5nIGJlZm9yZSBh'
    b'cHBlYXJpbmcgYXMgVVNCIGRyaXZlQQewUAWtSWdub3JlOiBzdG9wIHJlY29yZGluZyB3aG'
    b'VuIGJ1dHRvbiBpcyBwcmVzc2VkQQelUAWiU3RhcnQgcmVjb3JkaW5nIHdoZW4gVVNCIGNv'
    b'bm5lY3RlZEAj51ABgwv/f1AV3kVudGVyIGxvY2FsIHRpbWV6b25lJ3Mgb2Zmc2V0IGZyb2'
    b'0gVVRDIHRpbWUuIEltcGFjdHMgdGltZXN0YW1wIGluZm9ybWF0aW9uIG9uIHJlY29yZGlu'
    b'ZyBmaWxlcy5AL4BAGEIIUAWMRW5hYmxlIFdpLUZpUAGDGP9/UQGBAUAn7FABgxn/f1EBgQ'
    b'FBB6lQBaZVcGxvYWQgcmVjb3JkaW5nIHRvIGNsb3VkIGFmdGVyIGZpbmlzaFAVs1NlbGVj'
    b'dCB0byB1cGxvYWQgZWFjaCBmaWxlIHdoZW4gcmVjb3JkaW5nIGNvbXBsZXRlc0AnQLZQAY'
    b'Ma/39RAYEAQQejUAWgRGVsZXRlIGZyb20gc3RvcmFnZSBhZnRlciB1cGxvYWRQFUCCUmVj'
    b'b3JkaW5nIGZpbGUgd2lsbCBiZSBkZWxldGVkIGFmdGVyIGEgc3VjY2Vzc2Z1bCB1cGxvYW'
    b'QsIGFuZCBtb3ZlZCB0byBhIHN1YmZvbGRlciBsYWJlbGVkICdmYWlsZWQnIGluIHRoZSBj'
    b'YXNlIG9mIGEgZmFpbGVkIHVwbG9hZEAnQMJQAYMd/39RAYEAQQepUAWmS2VlcCB0aGUgV2'
    b'ktRmkgbW9kdWxlIG9uIGFuZCBjb25uZWN0ZWRQFUCISWYgY2hlY2tlZCwgdGhlIFdpLUZp'
    b'IG1vZHVsZSB3aWxsIHN0YXkgb24gYW5kIGNvbm5lY3RlZC4gIEFsbG93cyBmb3IgZmFzdG'
    b'VyIHVwbG9hZCB0aW1lcyB3aXRoIHRoZSBleHBlbnNlIG9mIGluY3JlYXNlZCBwb3dlciBj'
    b'b25zdW1wdGlvbkAPgEBPgEAoR3RQBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQAY'
    b'MS/39RAYEAUBVAj0RlbGF5IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZSwg'
    b'dGhlbiB3YWl0IGZvciB0cmlnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0ZX'
    b'Igd2FpdCBvciBvbiB0cmlnZ2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQeV'
    b'UAWSRGVsYXkgVGhlbiBUcmlnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0YX'
    b'J0IGF0IFRpbWVQAYMP/39QFdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdWF0'
    b'aW5nIG90aGVyIHRyaWdnZXIgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5nIG'
    b'RlbGF5QBGuUAWUUmVjb3JkaW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb25k'
    b'c0ARQJBQBZhSZWNvcmRpbmcgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25kc1'
    b'EhgwFRgFAV2FRpbWUgYmVmb3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZSBi'
    b'ZXR3ZWVuIHJlY29yZGluZ3MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlSZW'
    b'NvcmRpbmcgRmlsZSBTaXplIExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZXRy'
    b'aWdnZXJQAYMO/39QFLBDb25maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDExRk'
    b'Y3Rl09PW51bGxQFc9XaGVuIHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlciBy'
    b'ZWNvcmRpbmcgdGltZSBvciBzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQgZm'
    b'9yIEFsbCBTZW5zb3IgQ29uZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbmcg'
    b'd2lsbCBvbmx5IHN0YXJ0IHdoZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lzZS'
    b'ByZWNvcmRpbmcgd2lsbCBzdGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG1l'
    b'dEAYQQpQBZBQcmVzc3VyZSBUcmlnZ2VyUAGDBQAkUQCBAFAUmUNvbmZpZ1sweDAxRkYyNF'
    b'0gJiAxID09IDBAguVQBZVQcmVzc3VyZSBUcmlnZ2VyLCBMb3dQAYMDACRRAYMBX5BQJYJQ'
    b'YVEhgwHUwFAVs1NldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cm'
    b'Ugd2luZG93LkCC5lAFllByZXNzdXJlIFRyaWdnZXIsIEhpZ2hQAYMEACRRAYMBrbBQJYJQ'
    b'YVEhgwHUwFAVs1NldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cm'
    b'Ugd2luZG93LkAYQRdQBZNUZW1wZXJhdHVyZSBUcmlnZ2VyUAGDBQEkUQCBAFAUmUNvbmZp'
    b'Z1sweDAxRkYyNF0gJiAyID09IDBAY+pQBZhUZW1wZXJhdHVyZSBUcmlnZ2VyLCBMb3dQAY'
    b'MDASRRAoH2UCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0'
    b'c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQGPrUAWZVGVtcGVyYXR1cmUgVHJpZ2dlciwgSG'
    b'lnaFABgwQBJFECgSNQJYFDURKB2FEigVBQFbZTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hl'
    b'biBvdXRzaWRlIHRlbXBlcmF0dXJlIHdpbmRvdy5AGEGjUAWaQWNjZWxlcmF0aW9uIFRyaW'
    b'dnZXIgKDQwZylQAYMH/1BQFdRMb3cgcG93ZXIgdHJpZ2dlciB1c2luZyBkaWdpdGFsIGFj'
    b'Y2VsZXJvbWV0ZXIgd2l0aCBEQyByZXNwb25zZS4gTm8gcHJlLXRyaWdnZXIgZGF0YS5RAI'
    b'EAUBGDBf8IUBGDB/8gUCSAUBSUQ29uZmlnWzB4MWZmNTBdID09IDBAQ0CyUAWWQWNjZWxl'
    b'cmF0aW9uIFRocmVzaG9sZFABgwT/UFEBgiAAUgOIP0QAFAAUABRRI4hARAAAAAAAAFETgF'
    b'AV8kxvdyBwb3dlciB0cmlnZ2VyLiBUaGUgbWluaW11bSBhY2NlbGVyYXRpb24gKHBvc2l0'
    b'aXZlIG9yIG5lZ2F0aXZlKSB0byB0cmlnZ2VyIGEgcmVjb3JkaW5nLiBNdXN0IGJlIGdyZW'
    b'F0ZXIgdGhhbiAwLkAnxlABgwX/UFEBgQdBB5FQBY5YIEF4aXMgVHJpZ2dlckEHkVAFjlkg'
    b'QXhpcyBUcmlnZ2VyQQeRUAWOWiBBeGlzIFRyaWdnZXJAD4BAT4BAKEiCUAWMTWVhc3VyZW'
    b'1lbnRzQAhCBVAFlEFuYWxvZyBJbnB1dHMgKENoIDgpUCSAQCfKUQGBD1ABgwH/CFAVoEVu'
    b'YWJsZS9EaXNhYmxlIGFjY2VsZXJhdGlvbiBheGVzQQeaUAWTRW5hYmxlIE1pYyAoQ2ggOC'
    b'4zKVEBgQNAEUCaUAWLU2FtcGxlIFJhdGVQAYMC/whQFdhTZXQgbWFpbiBhY2NlbGVyYXRp'
    b'b24gc2FtcGxlIHJhdGUgZnJvbSAxMCB0byAyMDAwMCBIei4KSGlnaGVyIHJhdGUgcmVkdW'
    b'NlcyBiYXR0ZXJ5IGxpZmUuUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAwUQGCE4hQJYJIelER'
    b'gQpRIYJOIEARQPxQBaNPdmVycmlkZSBBbnRpYWxpYXNpbmcgRmlsdGVyIEN1dG9mZlABgw'
    b'j/CFEBggPoUCWCSHpRIYJhqFAVQKVEZWZhdWx0LCB3aGVuIG5vdCBjaGVja2VkLCBvZiBv'
    b'bmUtZmlmdGggc2FtcGxlIGZyZXF1ZW5jeS4gQ3V0LW9mZiBmcmVxdWVuY3kgaGFzIDI5JS'
    b'BhdHRlbnVhdGlvbi4KNXRoIG9yZGVyIEJ1dHRlcndvcnRoLiBTb21lIGF0dGVudWF0aW9u'
    b'IGF0IDYwJSBvZiBjdXQtb2ZmIGZyZXF1ZW5jeS5QFJVDb25maWdbMHgwMUZGMDhdID09ID'
    b'BACEEcUAWbNDBnIERDIEFjY2VsZXJhdGlvbiAoQ2ggODApQCefUAGDAf9QUQGBAUEHklAF'
    b'j0VuYWJsZSBBbGwgQXhlc0AXQNhQBYtTYW1wbGUgUmF0ZVABgwL/UFEBggH0UCWCSHpQFd'
    b'5Mb3cgcGFzcyBmaWx0ZXIgaXMgc2V0IHRvIDI1JSBvZiBkYXRhIHJhdGUsIHNvIGEgZGF0'
    b'YSByYXRlIG9mIDIwMDBIeiBoYXMgYSBMUCBmaWx0ZXIgYXQgNTAwIEh6UBSTbm90IENvbm'
    b'ZpZ1sweDFmZjUwXUEHhVEBgg+gQQeFUQGCB9BBB4VRAYID6EEHhVEBggH0QQeEUQGB+kEH'
    b'hFEBgX1BB4RRAYE/QQeEUQGBIEEHhFEBgRBACEEwUAWoQ29udHJvbCBQYWQgVGVtcGVyYX'
    b'R1cmUvUHJlc3N1cmUgKENoIDU5KUAnQJtRAYEDUAGDAf87UBWlRW5hYmxlL0Rpc2FibGUg'
    b'c2Vuc29ycyBvbiBjb250cm9sIHBhZEEHnFAFmUVuYWJsZSBQcmVzc3VyZSAoQ2ggNTkuMC'
    b'lBB59QBZxFbmFibGUgVGVtcGVyYXR1cmUgKENoIDU5LjEpQQelUAWiRW5hYmxlIFJlbGF0'
    b'aXZlIEh1bWlkaXR5IChDaCA1OS4yKUAR41AFi1NhbXBsZSBSYXRlUAGDAv87UBSUQ29uZm'
    b'lnWzB4MWZmM0JdID09IDBQFaRTZXQgc2FtcGxlIGZyZXF1ZW5jeSBmcm9tIDEgdG8gMTAg'
    b'SHpRAYEKUCWCSHpREYEBUSGBCkAIQc1QBZlJbmVydGlhbCBNZWFzdXJlbWVudCBVbml0QA'
    b'dAllAFkEFjcXVpc2l0aW9uIE1vZGVQAYMB/ytQJYEgUQGBEEEHilAFg09mZlEBgQBBB6hQ'
    b'BaFBYnNvbHV0ZSBPcmllbnRhdGlvbiAoUXVhdGVybmlvbilRAYEIQQeoUAWhUmVsYXRpdm'
    b'UgT3JpZW50YXRpb24gKFF1YXRlcm5pb24pUQGBEEEHj1AFiFJvdGF0aW9uUQGBAkABQIZQ'
    b'BZVPcmllbnRhdGlvbiBEYXRhIFJhdGVQAYMCCCtRAYFkUCWCSHpREYEBUSGByFAVvVNhbX'
    b'BsZSByYXRlIGZvciBkaXJlY3Rpb25hbCBkYXRhIGZyb20gMSB0byAyMDAgSHouIENoYW5u'
    b'ZWwgNjVQFJRDb25maWdbMHgwMUZGMkJdIDwgOEABQIlQBZJSb3RhdGlvbiBEYXRhIFJhdG'
    b'VQAYMCAitRAYFkUCWCSHpREYEBUSGByFAVvlNldCBzYW1wbGUgcmF0ZSBmb3Igcm90YXRp'
    b'b24gZGF0YSBmcm9tIDEgdG8gMjAwIEh6LiBDaGFubmVsIDQ3UBSZQ29uZmlnWzB4MDFGRj'
    b'JCXSAmIDIgPT0gMEAIQPFQBaVJbnRlcm5hbCBUZW1wZXJhdHVyZS9QcmVzc3VyZSAoQ2gg'
    b'MzYpQCdAxVEBgQNQAYMB/yRQFfdFbmFibGUvRGlzYWJsZSBzbG93ZXIgaW50ZXJuYWwgZW'
    b'52aXJvbm1lbnRhbCBzZW5zb3JzLiBOT1RFOiBUZW1wZXJhdHVyZSBpcyByZXF1aXJlZCBm'
    b'b3IgbWFpbiBhY2NlbGVyb21ldGVyIG1lYXN1cmVtZW50c0EHnFAFmUVuYWJsZSBQcmVzc3'
    b'VyZSAoQ2ggMzYuMClBB59QBZxFbmFibGUgVGVtcGVyYXR1cmUgKENoIDM2LjEpQAhAkFAF'
    b'oENvbnRyb2wgUGFkIExpZ2h0IFNlbnNvciAoQ2ggNzYpQCfqUAGDAf9MUBW+RW5hYmxlL0'
    b'Rpc2FibGUgbGlnaHQgc2Vuc29yIGF0IDQgSHogLSBOT1RFOiBESVNBQkxFRCBGT1IgQUxQ'
    b'SEFBB6BQBZ1FbmFibGUgTGlnaHQgU2Vuc29yIChDaCA3Ni4wKUAIQLJQBZNCTUcyNTAgR3'
    b'lybyAoQ2ggODQpQCefUAGDAf9UUQGBAEEHklAFj0VuYWJsZSBBbGwgQXhlc0AX91AFi1Nh'
    b'bXBsZSBSYXRlUAGDAv9UUQGCAZBQJYJIelAUk25vdCBDb25maWdbMHgxZmY1NF1BB4VRAY'
    b'IMgEEHhVEBggZAQQeFUQGCAyBBB4VRAYIBkEEHhFEBgchBB4RRAYFkQQeEUQGBMkEHhFEB'
    b'gRlBB4RRAYEMQA+AQE+ASgiASiiASkiA'
)



"""
Import and export of configuration data.

Note: User calibration and Wi-Fi setting are *not* included in the standard
import or export of configuration data. Calibration is specific to one
device, and for security reasons, Wi-Fi settings cannot be extracted from
the device.
"""

__all__ = ('exportConfig', 'importConfig')

from pathlib import Path
from typing import List, Optional, Union

from ebmlite import loadSchema, MasterElement

from .base import Recorder
from .exceptions import ConfigError
from .util import cleanProps


def deviceFromExport(export: Union[str, Path, MasterElement]) -> Recorder:
    """ Create a minimal virtual `Recorder` from exported config data. This
        virtual `Recorder` will have only the data relevant to configuration:
        no channels, sensors, calibration, et cetera.

        :param export: The name of an exported config file (``.xcg``), or
            EBML data containing an ``ExportedConfigurationData`` element.
        :return: A minimal 'virtual' `Recorder` instance.
    """
    if not isinstance(export, MasterElement):
        with open(export, 'rb') as f:
            export = loadSchema('mide_ide.xml').loads(f.read())

    if export[0].name == "ExportedConfigurationData":
        export = export[0]

    configData = None
    rawinfo = None
    configUi = None

    for el in export:
        if el.name == "RecorderConfigurationList":
            configData = el
        elif el.name == "RecordingProperties":
            rawinfo = el.getRaw()
        elif el.name == "ConfigUI":
            configUi = loadSchema('mide_config_ui.xml').loads(el.value)

    dev = Recorder(None, virtual=True, devinfo=rawinfo)
    dev._devinfo = None
    dev._source = export
    dev._configUi = configUi
    dev._configData = configData
    dev.getInfo()

    return dev


def exportConfig(device: Recorder,
                 filename: Union[str, Path],
                 unknown: bool = False,
                 defaults: bool = False) -> dict:
    """ Generate a configuration export file (``.xcg``). Writes the device's
        current information and configuration data by default.

        Note: User calibration and Wi-Fi setting are *not* included in
        exported configuration data. Calibration is specific to one
        device, and for security reasons, Wi-Fi settings cannot be
        extracted from the device.

        :param device: The device from which to export the config.
        :param filename: The name of the file to write.
        :param unknown: If `True`, include values read from the config
            file that did not correspond to known configuration items.
        :param defaults: If `True`, include config values that have not
            been explicitly set (i.e. still their default value).
    """
    config = device.config._makeConfig(unknown=unknown, defaults=defaults)
    configUi = device.config.getConfigUI()
    props = {'RecorderInfo': device.getInfo()}

    # Get contents; the outer element is added on encoding.
    config = config.get('RecorderConfigurationList', config)

    # Encode and write
    data = {'RecorderConfigurationList': config,
            'RecordingProperties': cleanProps(props),
            'ConfigUI': configUi.getRaw()}

    with open(filename, 'wb') as f:
        loadSchema('mide_ide.xml').encode(f, {'ExportedConfigurationData': data})

    return data


def importConfig(device: Recorder,
                 filename: Union[str, Path],
                 merge: bool = False,
                 exclude: Optional[List[int]] = None):
    """ Import configuration data from a ``.xcg`` file.

        Note: User calibration and Wi-Fi setting are *not* included in
        imported configuration data. Calibration is specific to one
        device, and for security reasons, Wi-Fi settings cannot be
        extracted from the device.

        :param device: The device to which to import the configuration data.
        :param filename: The name of an exported config file (``.xcg``).
        :param merge: If `True`, keep any device config values not
            explicitly set in the imported configuration data.
        :param exclude: An optional list of configuration IDs to ignore.
            These will neither be imported from the file, nor set to default
            if `merge` is `True`.
        :return:
    """
    exclude = tuple() if exclude is None else exclude
    imported = deviceFromExport(filename)
    for configId, item in device.config.items.items():
        if configId in exclude:
            continue
        if configId in imported.config.items:
            item.value = imported.config.items[configId].value
        elif not merge:
            item.value = None



"""
Default ConfigUI for S4-D40 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dSw0AoRZNQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABftQBZNSZW'
    b'NvcmRpbmcgRGlyZWN0b3J5UAGDFP9/UQSGUkVDT1JEUSSBEFAUk0NvbmZpZ1sweDE2RkY3'
    b'Rl09PTBQFblEaXJlY3RvcnkgZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5ncy4gMTYgY2hhcm'
    b'FjdGVyIG1heGltdW1ABUCjUAWVUmVjb3JkaW5nIEZpbGUgUHJlZml4UAGDFf9/USSBEFAV'
    b'6FByZWZpeCBmb3Igc2VyaWFsaXplZCByZWNvcmRpbmcgbmFtZXMuIElmIGJsYW5rLCBkZX'
    b'ZpY2Ugc2VyaWFsIG51bWJlciB3aWxsIGJlIHVzZWQuIDE2IGNoYXJhY3RlciBtYXhpbXVt'
    b'UBSTQ29uZmlnWzB4MTZGRjdGXT09MEAHQTxQBY9GaWxlIE5hbWUgU3R5bGVQAYMW/39RAY'
    b'EBUBVA/FRpbWUtQmFzZWQgZmlsZSBuYW1lcyB1c2UgYSBkaXJlY3RvcnkgYmFzZWQgb24g'
    b'dGhlIGRhdGUsIGFuZCBmaWxlIG5hbWUgYmFzZWQgb24gdGhlIHNlY29uZHMgc2luY2UgbW'
    b'lkbmlnaHQuClNlcmlhbGl6ZWQgZmlsZSBuYW1lcyB1c2UgdGhlIHNwZWNpZmllZCBkaXJl'
    b'Y3RvcnkgYW5kIGZpbGUgbmFtZSBwcmVmaXgsIGZvbGxvd2VkIGJ5IGEgbnVtYmVyLiAKTn'
    b'VtYmVyIGlzIHJlc2V0IHdoZW4gZGlyZWN0b3J5IG9yIHByZWZpeCBjaGFuZ2VzLkEHjVAF'
    b'ilRpbWUgQmFzZWRBB41QBYpTZXJpYWxpemVkQAdAqVAFi0J1dHRvbiBNb2RlUAGDEP9/UQ'
    b'GBAFAVv09uZSBTZWNvbmQgUHJlc3Mgd2lsbCBkaXNhYmxlIDMtc2Vjb25kIHByZXNzIGZv'
    b'ciBiYXR0ZXJ5IGNoZWNrLkEHl1AFlEltbWVkaWF0ZSBTdGFydC9TdG9wQQeiUAWfT25lIF'
    b'NlY29uZCBQcmVzcyBmb3IgU3RhcnQvU3RvcEEHjVAFilN0YXJ0LU9ubHlAF0EUUAWOUGx1'
    b'Zy1JbiBBY3Rpb25QAYMK/39RAYEBUBWtU2V0IGFjdGlvbiBmb3IgZGV2aWNlIHdoZW4gcG'
    b'x1Z2dlZCBpbiB0byBVU0IuQQe1UAWySW1tZWRpYXRlbHkgc3RvcCByZWNvcmRpbmcgYW5k'
    b'IGFwcGVhciBhcyBVU0IgZHJpdmVBB7NQBbBDb21wbGV0ZSByZWNvcmRpbmcgYmVmb3JlIG'
    b'FwcGVhcmluZyBhcyBVU0IgZHJpdmVBB7BQBa1JZ25vcmU6IHN0b3AgcmVjb3JkaW5nIHdo'
    b'ZW4gYnV0dG9uIGlzIHByZXNzZWRBB6VQBaJTdGFydCByZWNvcmRpbmcgd2hlbiBVU0IgY2'
    b'9ubmVjdGVkQCPnUAGDC/9/UBXeRW50ZXIgbG9jYWwgdGltZXpvbmUncyBvZmZzZXQgZnJv'
    b'bSBVVEMgdGltZS4gSW1wYWN0cyB0aW1lc3RhbXAgaW5mb3JtYXRpb24gb24gcmVjb3JkaW'
    b'5nIGZpbGVzLkAvgEAPgEBPgEAoR3RQBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQ'
    b'AYMS/39RAYEAUBVAj0RlbGF5IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZS'
    b'wgdGhlbiB3YWl0IGZvciB0cmlnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0'
    b'ZXIgd2FpdCBvciBvbiB0cmlnZ2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQ'
    b'eVUAWSRGVsYXkgVGhlbiBUcmlnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0'
    b'YXJ0IGF0IFRpbWVQAYMP/39QFdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdW'
    b'F0aW5nIG90aGVyIHRyaWdnZXIgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5n'
    b'IGRlbGF5QBGuUAWUUmVjb3JkaW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb2'
    b'5kc0ARQJBQBZhSZWNvcmRpbmcgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25k'
    b'c1EhgwFRgFAV2FRpbWUgYmVmb3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZS'
    b'BiZXR3ZWVuIHJlY29yZGluZ3MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlS'
    b'ZWNvcmRpbmcgRmlsZSBTaXplIExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZX'
    b'RyaWdnZXJQAYMO/39QFLBDb25maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDEx'
    b'RkY3Rl09PW51bGxQFc9XaGVuIHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlci'
    b'ByZWNvcmRpbmcgdGltZSBvciBzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQg'
    b'Zm9yIEFsbCBTZW5zb3IgQ29uZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbm'
    b'cgd2lsbCBvbmx5IHN0YXJ0IHdoZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lz'
    b'ZSByZWNvcmRpbmcgd2lsbCBzdGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG'
    b'1ldEAYQQpQBZBQcmVzc3VyZSBUcmlnZ2VyUAGDBQAkUQCBAFAUmUNvbmZpZ1sweDAxRkYy'
    b'NF0gJiAxID09IDBAguVQBZVQcmVzc3VyZSBUcmlnZ2VyLCBMb3dQAYMDACRRAYMBX5BQJY'
    b'JQYVEhgwHUwFAVs1NldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgcHJlc3N1'
    b'cmUgd2luZG93LkCC5lAFllByZXNzdXJlIFRyaWdnZXIsIEhpZ2hQAYMEACRRAYMBrbBQJY'
    b'JQYVEhgwHUwFAVs1NldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgcHJlc3N1'
    b'cmUgd2luZG93LkAYQRdQBZNUZW1wZXJhdHVyZSBUcmlnZ2VyUAGDBQEkUQCBAFAUmUNvbm'
    b'ZpZ1sweDAxRkYyNF0gJiAyID09IDBAY+pQBZhUZW1wZXJhdHVyZSBUcmlnZ2VyLCBMb3dQ'
    b'AYMDASRRAoH2UCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3'
    b'V0c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQGPrUAWZVGVtcGVyYXR1cmUgVHJpZ2dlciwg'
    b'SGlnaFABgwQBJFECgSNQJYFDURKB2FEigVBQFbZTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2'
    b'hlbiBvdXRzaWRlIHRlbXBlcmF0dXJlIHdpbmRvdy5AGEGjUAWaQWNjZWxlcmF0aW9uIFRy'
    b'aWdnZXIgKDQwZylQAYMH/1BQFdRMb3cgcG93ZXIgdHJpZ2dlciB1c2luZyBkaWdpdGFsIG'
    b'FjY2VsZXJvbWV0ZXIgd2l0aCBEQyByZXNwb25zZS4gTm8gcHJlLXRyaWdnZXIgZGF0YS5R'
    b'AIEAUBGDBf8IUBGDB/8gUCSAUBSUQ29uZmlnWzB4MWZmNTBdID09IDBAQ0CyUAWWQWNjZW'
    b'xlcmF0aW9uIFRocmVzaG9sZFABgwT/UFEBgiAAUgOIP0QAFAAUABRRI4hARAAAAAAAAFET'
    b'gFAV8kxvdyBwb3dlciB0cmlnZ2VyLiBUaGUgbWluaW11bSBhY2NlbGVyYXRpb24gKHBvc2'
    b'l0aXZlIG9yIG5lZ2F0aXZlKSB0byB0cmlnZ2VyIGEgcmVjb3JkaW5nLiBNdXN0IGJlIGdy'
    b'ZWF0ZXIgdGhhbiAwLkAnxlABgwX/UFEBgQdBB5FQBY5YIEF4aXMgVHJpZ2dlckEHkVAFjl'
    b'kgQXhpcyBUcmlnZ2VyQQeRUAWOWiBBeGlzIFRyaWdnZXJAD4BAT4BAKEWnUAWMTWVhc3Vy'
    b'ZW1lbnRzQAhBHFAFmzQwZyBEQyBBY2NlbGVyYXRpb24gKENoIDgwKUAnn1ABgwH/UFEBgQ'
    b'FBB5JQBY9FbmFibGUgQWxsIEF4ZXNAF0DYUAWLU2FtcGxlIFJhdGVQAYMC/1BRAYIB9FAl'
    b'gkh6UBXeTG93IHBhc3MgZmlsdGVyIGlzIHNldCB0byAyNSUgb2YgZGF0YSByYXRlLCBzby'
    b'BhIGRhdGEgcmF0ZSBvZiAyMDAwSHogaGFzIGEgTFAgZmlsdGVyIGF0IDUwMCBIelAUk25v'
    b'dCBDb25maWdbMHgxZmY1MF1BB4VRAYIPoEEHhVEBggfQQQeFUQGCA+hBB4VRAYIB9EEHhF'
    b'EBgfpBB4RRAYF9QQeEUQGBP0EHhFEBgSBBB4RRAYEQQAhBMFAFqENvbnRyb2wgUGFkIFRl'
    b'bXBlcmF0dXJlL1ByZXNzdXJlIChDaCA1OSlAJ0CbUQGBA1ABgwH/O1AVpUVuYWJsZS9EaX'
    b'NhYmxlIHNlbnNvcnMgb24gY29udHJvbCBwYWRBB5xQBZlFbmFibGUgUHJlc3N1cmUgKENo'
    b'IDU5LjApQQefUAWcRW5hYmxlIFRlbXBlcmF0dXJlIChDaCA1OS4xKUEHpVAFokVuYWJsZS'
    b'BSZWxhdGl2ZSBIdW1pZGl0eSAoQ2ggNTkuMilAEeNQBYtTYW1wbGUgUmF0ZVABgwL/O1AU'
    b'lENvbmZpZ1sweDFmZjNCXSA9PSAwUBWkU2V0IHNhbXBsZSBmcmVxdWVuY3kgZnJvbSAxIH'
    b'RvIDEwIEh6UQGBClAlgkh6URGBAVEhgQpACEHNUAWZSW5lcnRpYWwgTWVhc3VyZW1lbnQg'
    b'VW5pdEAHQJZQBZBBY3F1aXNpdGlvbiBNb2RlUAGDAf8rUCWBIFEBgRBBB4pQBYNPZmZRAY'
    b'EAQQeoUAWhQWJzb2x1dGUgT3JpZW50YXRpb24gKFF1YXRlcm5pb24pUQGBCEEHqFAFoVJl'
    b'bGF0aXZlIE9yaWVudGF0aW9uIChRdWF0ZXJuaW9uKVEBgRBBB49QBYhSb3RhdGlvblEBgQ'
    b'JAAUCGUAWVT3JpZW50YXRpb24gRGF0YSBSYXRlUAGDAggrUQGBZFAlgkh6URGBAVEhgchQ'
    b'Fb1TYW1wbGUgcmF0ZSBmb3IgZGlyZWN0aW9uYWwgZGF0YSBmcm9tIDEgdG8gMjAwIEh6Li'
    b'BDaGFubmVsIDY1UBSUQ29uZmlnWzB4MDFGRjJCXSA8IDhAAUCJUAWSUm90YXRpb24gRGF0'
    b'YSBSYXRlUAGDAgIrUQGBZFAlgkh6URGBAVEhgchQFb5TZXQgc2FtcGxlIHJhdGUgZm9yIH'
    b'JvdGF0aW9uIGRhdGEgZnJvbSAxIHRvIDIwMCBIei4gQ2hhbm5lbCA0N1AUmUNvbmZpZ1sw'
    b'eDAxRkYyQl0gJiAyID09IDBACEDxUAWlSW50ZXJuYWwgVGVtcGVyYXR1cmUvUHJlc3N1cm'
    b'UgKENoIDM2KUAnQMVRAYEDUAGDAf8kUBX3RW5hYmxlL0Rpc2FibGUgc2xvd2VyIGludGVy'
    b'bmFsIGVudmlyb25tZW50YWwgc2Vuc29ycy4gTk9URTogVGVtcGVyYXR1cmUgaXMgcmVxdW'
    b'lyZWQgZm9yIG1haW4gYWNjZWxlcm9tZXRlciBtZWFzdXJlbWVudHNBB5xQBZlFbmFibGUg'
    b'UHJlc3N1cmUgKENoIDM2LjApQQefUAWcRW5hYmxlIFRlbXBlcmF0dXJlIChDaCAzNi4xKU'
    b'AI9VAFoENvbnRyb2wgUGFkIExpZ2h0IFNlbnNvciAoQ2ggNzYpQCfPUAGDAf9MUBWjRW5h'
    b'YmxlL0Rpc2FibGUgbGlnaHQgc2Vuc29yIGF0IDQgSHpBB6BQBZ1FbmFibGUgTGlnaHQgU2'
    b'Vuc29yIChDaCA3Ni4wKUAPgEBPgEoIgEoogEpIgA=='
)



""" 
ebmlite.schemata: Submodule containing the default schema XML files. It doesn't
contain any code; it's just for storage. Making it a module makes it easy to
find. 

For the definition of the `Schema` class, see `ebmlite.core`.
"""



"""
Initial 'sanity check' identification and instantiation tests. Perform early.
"""

import os.path
from glob import glob
import idelib.importer
import pytest

import endaq.device

from . import fake_recorders

# Create parameters, mainly to provide an ID, making the results more readable
RECORDER_PATHS = [pytest.param(path, id=os.path.basename(path))
                  for path in fake_recorders.RECORDER_PATHS]


TEST_ROOT = os.path.dirname(__file__)
IDE_FILES = glob(os.path.join(TEST_ROOT, 'recordings', '*.IDE'))


@pytest.mark.parametrize("path", RECORDER_PATHS)
def test_basic_identification(path):
    """ Basic test of recorder identification. All test 'fake recorders'
        should be identified.
    """
    assert endaq.device.isRecorder(path, strict=False)


def test_identification_negative():
    """ Test a non-recorder path to verify it isn't misidentified.
    """
    # Using the test file's directory because it's handy.
    assert not endaq.device.isRecorder(os.path.dirname(__file__))


@pytest.mark.parametrize("path", RECORDER_PATHS)
def test_basic_instantiation(path):
    """ Basic test of recorder instantiation. All test 'fake recorders'
        should instantiate `Recorder` subclasses.
    """
    endaq.device.RECORDERS.clear()  # Clear any cached devices, just to be safe
    dev = endaq.device.getRecorder(path, strict=False)
    assert dev.path == path
    assert dev.partNumber == os.path.basename(path).partition('_')[0]


@pytest.mark.parametrize("path", RECORDER_PATHS)
def test_basic_caching(path):
    """ Test that previously created recorders are cached and reused.
    """
    dev = endaq.device.getRecorder(path, strict=False)
    dev2 = endaq.device.getRecorder(path, strict=False)
    assert dev is dev2

    # Clear cache and verify new recorders are created
    endaq.device.RECORDERS.clear()
    dev3 = endaq.device.getRecorder(path, strict=False)
    assert dev is not dev3


def test_getDevices():
    """ Test of `getDevices()`, comparing found device paths to the list of
        fake recorder directories.
    """
    endaq.device.RECORDERS.clear()
    devs = endaq.device.getDevices(paths=fake_recorders.RECORDER_PATHS,
                                   strict=False,
                                   unmounted=False)
    assert sorted(dev.path for dev in devs) == sorted(fake_recorders.RECORDER_PATHS)

    # Just one path provided should return just one device
    devs = endaq.device.getDevices(paths=fake_recorders.RECORDER_PATHS[0],
                                   strict=False,
                                   unmounted=False)
    assert len(devs) == 1


@pytest.mark.parametrize("path", RECORDER_PATHS)
def test_onRecorder(path):
    """ Test checking if a file is on a recorder (its path corresponds
        to a recorder's path).
    """
    endaq.device.RECORDERS.clear()
    dev = endaq.device.getRecorder(path, strict=False)
    assert endaq.device.onRecorder(dev.infoFile, strict=False)


@pytest.mark.parametrize("path", RECORDER_PATHS)
def test_findDevice(path):
    """ Test finding recorders by serial number.
    """
    with pytest.raises(ValueError):
        # ValueError: Either a serial number or chip ID is required
        endaq.device.findDevice(paths=fake_recorders.RECORDER_PATHS,
                                strict=False)

    with pytest.raises(ValueError):
        # ValueError: Either a serial number or chip ID is required, not both
        endaq.device.findDevice(sn=1234,
                                chipId=0xabcd,
                                paths=fake_recorders.RECORDER_PATHS,
                                strict=False)

    # Nonexistent serial number
    assert not endaq.device.findDevice(sn=-1,
                                       paths=fake_recorders.RECORDER_PATHS,
                                       strict=False)

    dev = endaq.device.getRecorder(path, strict=False)
    endaq.device.RECORDERS.clear()

    assert endaq.device.findDevice(dev.serialInt,
                                   paths=fake_recorders.RECORDER_PATHS,
                                   strict=False)
    assert endaq.device.findDevice(sn=dev.serialInt,
                                   paths=fake_recorders.RECORDER_PATHS,
                                   strict=False)
    assert endaq.device.findDevice(sn=dev.serial,
                                   paths=fake_recorders.RECORDER_PATHS,
                                   strict=False)

    if dev.chipId:
        # Very old HW/FW does not report a chip ID
        assert endaq.device.findDevice(chipId=dev.chipId,
                                       paths=fake_recorders.RECORDER_PATHS,
                                       strict=False)


@pytest.mark.parametrize("filename", IDE_FILES)
def test_fromRecording(filename):
    """ Test instantiation from an IDE file.
    """
    doc = idelib.importer.importFile(filename, quiet=True)
    dev = endaq.device.fromRecording(doc)

    assert dev.partNumber in filename

    # Sanity checks: Make sure channels/sensors/transforms were created
    # TODO: More detailed check of contents, and/or basic type checks?
    assert dev.channels
    assert dev.sensors
    assert dev.transforms



"""
'Strategy' classes for reading device information over different media
(e.g., file-based or through the command interface). These are similar to
`CommandInterface` and `ConfigInterface` in that regard, but these are not
accessed directly.
"""

from abc import ABC, abstractmethod
import logging
import os.path
import struct
from typing import Optional, Tuple, Union, TYPE_CHECKING

from .types import Drive, Filename
from .command_interfaces import SerialCommandInterface
from . import util

logger = logging.getLogger('endaq.device')

if TYPE_CHECKING:
    from .base import Recorder


# ===========================================================================
#
# ===========================================================================

class DeviceInfo(ABC):
    """
    An abstract mechanism for retrieving and setting device information. Its
    methods only read and write the raw binary; any parsing and/or is done by
    the caller.
    """

    @classmethod
    @abstractmethod
    def hasInterface(cls, device: "Recorder") -> bool:
        """ Determine if a device supports this info-accessing type.

            :param device: The recorder to check.
            :return: `True` if the device supports the interface.
        """
        raise NotImplementedError


    @classmethod
    @abstractmethod
    def readDevinfo(cls, path: Filename, info=None) -> Union[None, bytearray, bytes]:
        """ Calculate the device's hash. Separated from `__hash__()` so it
            can be used by `getDevices()` to find known recorders.

            :param path: The device's filesystem path.
            :param info: The contents of the device's `DEVINFO` file, if
                previously loaded. For future caching optimization.
        """
        raise NotImplementedError


    @abstractmethod
    def readManifest(self) \
            -> Tuple[Union[None, bytearray, bytes], Union[None, bytearray, bytes], Union[None, bytearray, bytes]]:
        """ Read the device's manifest data. The data is a superset of the
            information returned by `getInfo()`.
        """
        raise NotImplementedError


    @abstractmethod
    def readUserCalibration(self) -> Union[None, bytearray, bytes]:
        """ Get the recorder's user-defined calibration data as a dictionary
            of parameters.
        """
        raise NotImplementedError


    @abstractmethod
    def writeUserCal(self,
                     caldata: Union[None, bytearray, bytes]):
        """ Write user calibration to the device.

            :param caldata: The raw binary of an EBML `<CalibrationList>`
                element..
        """
        raise NotImplementedError


# ===========================================================================
#
# ===========================================================================

class FileDeviceInfo(DeviceInfo):
    """
    A mechanism for retrieving device information from files (mostly
    firmware-generated) on the device. Its methods only read and write the raw
    binary; any parsing and/or encoding is done by the caller.
    """

    _INFO_FILE = os.path.join("SYSTEM", "DEV", "DEVINFO")


    def __init__(self, device: 'Recorder', **_kwargs):
        self.device = device
    

    @classmethod
    def hasInterface(cls, device: "Recorder") -> bool:
        """ Determine if a device supports this `DeviceInfo` type.

            :param device: The recorder to check.
            :return: `True` if the device supports the interface.
        """
        if device.isVirtual:
            return False

        return device.path and os.path.exists(device.infoFile)


    @classmethod
    def readDevinfo(cls,
                    path: Filename,
                    info: Union[None, bytearray, bytes] = None) -> Union[None, bytearray, bytes]:
        """ Retrieve a DEVINFO data. Calculate the device's hash. Separated from `__hash__()` so it
            can be used by `getDevices()` to find known recorders.

            :param path: The device's filesystem path.
            :param info: The contents of the device's `DEVINFO` file, if
                previously loaded. For future caching optimization.
        """
        if path and not info:
            path = os.path.realpath(path.path if isinstance(path, Drive) else path)
            infoFile = os.path.join(path, cls._INFO_FILE)
            if os.path.exists(infoFile):
                with open(infoFile, 'rb') as f:
                    info = f.read()

        return info


    def _readUserpage(self) \
            -> Tuple[Union[None, bytearray, bytes], Union[None, bytearray, bytes], Union[None, bytearray, bytes]]:
        """ Read the device's manifest data from the EFM32 'userpage'. The
            data is a superset of the information returned by `getInfo()`.
            Factory calibration and recorder properties are also read
            and cached, since one or both are in the userpage.
        """
        if self.device._manifest is not None:
            return self.device._manifest

        # Recombine all the 'user page' files
        data = bytearray()
        for i in range(4):
            filename = os.path.join(self.device.path, self.device._USERPAGE_FILE % i)
            with open(filename, 'rb') as fs:
                data.extend(fs.read())

        (manOffset, manSize,
         calOffset, calSize,
         propOffset, propSize) = struct.unpack_from("<HHHHHH", data)

        manData = data[manOffset:manOffset + manSize]
        calData = data[calOffset:calOffset + calSize]

        # _propData is read and cached here but parsed in `getSensors()`.
        # New devices use a dynamically-generated properties file, which
        # overrides any property data in the USERPAGE.
        if os.path.exists(self.device.recpropFile):
            with open(self.device.recpropFile, 'rb') as f:
                propData = f.read()
        else:
            # Zero offset means no property data (very old devices). For new
            # devices, a size of 1 also means no data (it's a null byte).
            propSize = 0 if (propOffset == 0 or propSize <= 1) else propSize
            propData = data[propOffset:propOffset + propSize]

        return manData, calData, propData


    def _readManifest(self) \
            -> Tuple[Union[None, bytearray, bytes], Union[None, bytearray, bytes], Union[None, bytearray, bytes]]:
        """ Read the device's manifest data from the 'MANIFEST' file. The
            data is a superset of the information returned by `getInfo()`.

            Factory calibration and recorder properties are also read and
            cached for backwards compatibility, since both are in the older
            devices' EFM32 'userpage'.
        """
        manFile = os.path.join(self.device.path, self.device._MANIFEST_FILE)
        calFile = os.path.join(self.device.path, self.device._SYSCAL_FILE)

        manData = calData = propData = None

        try:
            with open(manFile, 'rb') as f:
                manData = f.read()
        except (FileNotFoundError, AttributeError) as err:
            logger.debug(f"Possibly-allowed exception when reading {manFile}: {err!r}")

        try:
            with open(calFile, 'rb') as f:
                calData = f.read()
        except (FileNotFoundError, AttributeError) as err:
            logger.debug(f"Possibly-allowed exception when reading {calFile}: {err!r}")

        try:
            # _propData is read and cached here but parsed in `getSensors()`.
            # Old EFM32 recorders stored this w/ the manifest in the USERPAGE.
            with open(self.device.recpropFile, 'rb') as f:
                propData = f.read()
        except (FileNotFoundError, AttributeError) as err:
            logger.debug("Possibly-allowed exception when reading "
                         f"{self.device.recpropFile}: {err!r}")

        return manData, calData, propData


    def readManifest(self) \
            -> Tuple[Union[None, bytearray, bytes], Union[None, bytearray, bytes], Union[None, bytearray, bytes]]:
        """ Read the device's manifest data. The data is a superset of the
            information returned by `getInfo()`.
        """
        if os.path.exists(os.path.join(self.device.path, self.device._USERPAGE_FILE % 0)):
            return  self._readUserpage()
        elif os.path.exists(os.path.join(self.device.path, self.device._MANIFEST_FILE)):
            return self._readManifest()

        return None, None, None


    def readUserCalibration(self) -> Union[None, bytearray, bytes]:
        """ Get the recorder's user-defined calibration data as a dictionary
            of parameters.
        """
        if not os.path.isfile(self.device.userCalFile):
            return None

        with open(self.device.userCalFile, 'rb') as f:
            return f.read()


    def writeUserCal(self,
                     caldata: Union[None, bytearray, bytes]):
        """ Write user calibration to the device.

            :param caldata: The raw binary of an EBML `<CalibrationList>`
                element..
        """
        if caldata:
            try:
                util.makeBackup(self.device.userCalFile)
                with open(self.device.userCalFile, 'wb') as f:
                    f.write(caldata)
            except Exception:
                util.restoreBackup(self.device.userCalFile)
                raise


# ===========================================================================
#
# ===========================================================================

class SerialDeviceInfo(DeviceInfo):
    """
    A mechanism for retrieving device information from a remote device connected
    via a serial command interface. Its methods only read and write the raw
    binary; any parsing and/or encoding is done by the caller.
    """

    def __init__(self, device: 'Recorder', **_kwargs):
        self.device = device


    @classmethod
    def hasInterface(cls, device: "Recorder") -> bool:
        """ Determine if a device supports this `DeviceInfo` type.

            :param device: The recorder to check.
            :return: `True` if the device supports the interface.
        """
        if device.isVirtual or FileDeviceInfo.hasInterface(device):
            return False

        # return isinstance(device.command, SerialCommandInterface)
        return SerialCommandInterface.hasInterface(device)


    @classmethod
    def readDevinfo(cls,
                    path: Union[Filename, "Recorder"],
                    info: Optional[bytes] = None) -> Optional[bytes]:
        """ Retrieve a DEVINFO data.

            :param path: The device's filesystem path.
            :param info: The contents of the device's `DEVINFO` file, if
                previously loaded. For future caching optimization.
        """
        if isinstance(info, bytearray):
            info = bytes(info)
        if isinstance(getattr(path, 'command', None), SerialCommandInterface):
            return info or path.command._getInfo(0)
        return info


    def readManifest(self) \
            -> Tuple[Optional[bytes], Optional[bytes], Optional[bytes]]:
        """ Read the device's manifest data from the 'MANIFEST' file. The
            data is a superset of the information returned by `getInfo()`.

            Factory calibration and recorder properties are also read and
            cached for backwards compatibility, since both are in the older
            devices' EFM32 'userpage'.
        """
        manData = self.device.command._getInfo(3) or None
        calData = self.device.command._getInfo(4) or None
        propData = self.device.command._getInfo(1) or None

        return manData, calData, propData


    def readUserCalibration(self) -> Union[None, bytearray, bytes]:
        """ Get the recorder's user-defined calibration data as a dictionary
            of parameters.
        """
        return self.device.command._getInfo(6) or None


    def writeUserCal(self,
                     caldata: Union[None, bytearray, bytes]):
        """ Write user calibration to the device.

            :param caldata: The raw binary of an EBML `<CalibrationList>`
                element..
        """
        caldata = caldata or b''
        self.device.command._setInfo(6, caldata)


# ===========================================================================
#
# ===========================================================================

class MQTTDeviceInfo(SerialDeviceInfo):
    """
    A mechanism for retrieving device information from a remote device connected
    via MQTT. Its methods only read and write the raw binary; any parsing and/or
    encoding is done by the caller.
    """

    # Very similar to SerialDeviceInfo, in that they get data via the command
    # interface.
    # XXX: PLACEHOLDER!
    def __init__(self, device: 'Recorder', **_kwargs):
        logger.warning('MQTTDeviceInfo not implemented!', NotImplemented)
        super().__init__(device, **_kwargs)


    @classmethod
    def hasInterface(cls, device: "Recorder") -> bool:
        """ Determine if a device supports this `DeviceInfo` type.

            Note: it's likely that this interface will be instantiated
            by the mechanism that instantiates MQTT devices, so this
            won't be used.

            :param device: The recorder to check.
            :return: `True` if the device supports the interface.
        """
        if device.isVirtual:
            return False

        # TODO: Better mechanism for identifying MQTT devices
        return device.path and device.path.lower().startswith('mqtt')


# ===========================================================================
#
# ===========================================================================

#: A list of all `DeviceInfo` types, used when finding a device's  interface.
#  New interface types defined elsewhere should append/insert themselves into
#  this list.
INTERFACES = [FileDeviceInfo,
              SerialDeviceInfo,
              MQTTDeviceInfo]



"""
Default CONFIG.UI data, for devices (mostly virtual) that do not have it.
CONFIG.UI data defines the configuration GUI layout, as well as provides
information for converting values when reading and writing configuration
data.
"""

from importlib import import_module
import logging
from typing import Union

logger = logging.getLogger('endaq.device')

# Special-case mapping of non-standard part numbers to default ConfigUI
# modules (special/custom/OEM builds, etc.)
# FUTURE: Add pattern matching (regex/glob)?
SPECIAL_CASES = {
    'S3-D16': 'LOG-0003-016G',
    'SF-DR4-01': 'W8-R2000D40',
    'SF-DR4-02': 'Sx-E2000D40',
    'SF-DR4-03': 'W8-R2000D40',
    'SF-DR4-04': 'Sx-E100D40',
}


def _getConfigUI(name: str) -> Union[bytes, None]:
    """ Get the default ConfigUI data from a module.

        :param name: The name of the module (typically matches a device's
            part number).
        :return: The raw binary ConfigUI EBML if there is a matching module,
            or `None`.
    """
    try:
        name = name.replace('-', '_')
        mod = import_module("." + name, __package__)
        return mod.DEFAULT_CONFIG_UI
    except ImportError:
        return None


def _getGenericName(pn: str) -> str:
    """ Get a potential module name for a product without a unique ConfigUI
        file for its part number. This is typically a 'generic' one, or
        a special case device name.

        :param pn: A product part number without a directly corresponding
            default ConfigUI module.
        :return: An alternative part number that may (or may not) have a
            corresponding default ConfigUI module.
    """
    if pn in SPECIAL_CASES:
        return SPECIAL_CASES[pn]

    elif pn.startswith('LOG-'):
        # For ancient Slam Stick X recorders without digital accel
        return pn + "_DC"

    family, sep, model = pn.partition('-')
    return "{}x{}{}".format(family[0], sep, model)


def getDefaultConfigUI(device) -> Union[str, None]:
    """ Attempt to find canned 'default' ConfigUI file for the device,
        based on its part number.

        :param device: The device in need of default ConfigUI data.
        :return: The raw binary ConfigUI EBML if there is a corresponding
            module, or `None`.
    """
    # FUTURE: Also have default variants based on HwRev and/or FwRev?
    uiName = device.partNumber
    ui = _getConfigUI(uiName)
    if not ui:
        ui = _getConfigUI(_getGenericName(uiName))

    if not ui:
        logger.warning("Could not find default ConfigUI for {}, using default".format(device.partNumber))
        ui = _getConfigUI('default')

    return ui



"""
Default ConfigUI for S2-D40D200 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dV4kAoRjBQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABUCZUAWVQ3'
    b'VzdG9tIFJlY29yZGluZyBUYWdzUAGDF/9/UQSAUSSBQFAV8UtleXdvcmRzIGZvciBvcmdh'
    b'bml6aW5nIGFuZCBpZGVudGlmeWluZyByZWNvcmRpbmdzLCBzZXBhcmF0ZWQgYnkgY29tbW'
    b'Egb3Igc3BhY2UuIExpbWl0IGlzIDIwIHRhZ3MsIDY0IGNoYXJhY3RlcnMuQAX7UAWTUmVj'
    b'b3JkaW5nIERpcmVjdG9yeVABgxT/f1EEhlJFQ09SRFEkgRBQFJNDb25maWdbMHgxNkZGN0'
    b'ZdPT0wUBW5RGlyZWN0b3J5IGZvciBzZXJpYWxpemVkIHJlY29yZGluZ3MuIDE2IGNoYXJh'
    b'Y3RlciBtYXhpbXVtQAVAo1AFlVJlY29yZGluZyBGaWxlIFByZWZpeFABgxX/f1EkgRBQFe'
    b'hQcmVmaXggZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5nIG5hbWVzLiBJZiBibGFuaywgZGV2'
    b'aWNlIHNlcmlhbCBudW1iZXIgd2lsbCBiZSB1c2VkLiAxNiBjaGFyYWN0ZXIgbWF4aW11bV'
    b'AUk0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGB'
    b'AVAVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIH'
    b'RoZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1p'
    b'ZG5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZW'
    b'N0b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51'
    b'bWJlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBY'
    b'pUaW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQKlQBYtCdXR0b24gTW9kZVABgxD/f1EB'
    b'gQBQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3'
    b'IgYmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBT'
    b'ZWNvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BBB41QBYpTdGFydC1Pbmx5QBdBFFAFjlBsdW'
    b'ctSW4gQWN0aW9uUAGDCv9/UQGBAVAVrVNldCBhY3Rpb24gZm9yIGRldmljZSB3aGVuIHBs'
    b'dWdnZWQgaW4gdG8gVVNCLkEHtVAFskltbWVkaWF0ZWx5IHN0b3AgcmVjb3JkaW5nIGFuZC'
    b'BhcHBlYXIgYXMgVVNCIGRyaXZlQQezUAWwQ29tcGxldGUgcmVjb3JkaW5nIGJlZm9yZSBh'
    b'cHBlYXJpbmcgYXMgVVNCIGRyaXZlQQewUAWtSWdub3JlOiBzdG9wIHJlY29yZGluZyB3aG'
    b'VuIGJ1dHRvbiBpcyBwcmVzc2VkQQelUAWiU3RhcnQgcmVjb3JkaW5nIHdoZW4gVVNCIGNv'
    b'bm5lY3RlZEAj51ABgwv/f1AV3kVudGVyIGxvY2FsIHRpbWV6b25lJ3Mgb2Zmc2V0IGZyb2'
    b'0gVVRDIHRpbWUuIEltcGFjdHMgdGltZXN0YW1wIGluZm9ybWF0aW9uIG9uIHJlY29yZGlu'
    b'ZyBmaWxlcy5AL4BAD4BAT4BAKEkbUAWIVHJpZ2dlcnNAB0DaUAWMVHJpZ2dlciBNb2RlUA'
    b'GDEv9/UQGBAFAVQI9EZWxheSBUaGVuIFRyaWdnZXI6IFdhaXQgc3BlY2lmaWVkIHRpbWUs'
    b'IHRoZW4gd2FpdCBmb3IgdHJpZ2dlcgpEZWxheSBPciBUcmlnZ2VyOiBBY3F1aXJlIGFmdG'
    b'VyIHdhaXQgb3Igb24gdHJpZ2dlci4gQmF0dGVyeSBsaWZlIG1heSBiZSByZWR1Y2VkLkEH'
    b'lVAFkkRlbGF5IFRoZW4gVHJpZ2dlckEHk1AFkERlbGF5IE9yIFRyaWdnZXJAMvRQBY1TdG'
    b'FydCBhdCBUaW1lUAGDD/9/UBXbV2FpdCB1bnRpbCBzZXQgdGltZSBiZWZvcmUgZXZhbHVh'
    b'dGluZyBvdGhlciB0cmlnZ2VyIGNvbmRpdGlvbnMsIGluY2x1ZGluZyBhIHJlY29yZGluZy'
    b'BkZWxheUARrlAFlFJlY29yZGluZyBUaW1lIExpbWl0UAGDDf9/USGE////8FAlh1NlY29u'
    b'ZHNAEUCQUAWYUmVjb3JkaW5nIERlbGF5L0ludGVydmFsUAGDDP9/UQGBAFAlh1NlY29uZH'
    b'NRIYMBUYBQFdhUaW1lIGJlZm9yZSBzdGFydGluZyBhIHJlY29yZGluZywgYW5kIHRpbWUg'
    b'YmV0d2VlbiByZWNvcmRpbmdzIGlmICdSZXRyaWdnZXInIGlzIGNoZWNrZWQuQBGvUAWZUm'
    b'Vjb3JkaW5nIEZpbGUgU2l6ZSBMaW1pdFABgxH/f1EhhP////BQJYNLaUJAEECXUAWJUmV0'
    b'cmlnZ2VyUAGDDv9/UBSwQ29uZmlnWzB4REZGN0ZdPT1udWxsIGFuZCBDb25maWdbMHgxMU'
    b'ZGN0ZdPT1udWxsUBXPV2hlbiBzZXQsIFNsYW0gU3RpY2sgd2lsbCByZS1hcm0gYWZ0ZXIg'
    b'cmVjb3JkaW5nIHRpbWUgb3Igc2l6ZSBsaW1pdCBpcyByZWFjaGVkLkAQQKhQBZ5XYWl0IG'
    b'ZvciBBbGwgU2Vuc29yIENvbmRpdGlvbnNQAYMT/39QFf5XaGVuIHNldCwgcmVjb3JkaW5n'
    b'IHdpbGwgb25seSBzdGFydCB3aGVuIGFsbCB0cmlnZ2VycyBhcmUgbWV0LCBvdGhlcndpc2'
    b'UgcmVjb3JkaW5nIHdpbGwgc3RhcnQgd2hlbiBhbnkgc2Vuc29yIGNvbmRpdGlvbiBpcyBt'
    b'ZXRAGEEKUAWQUHJlc3N1cmUgVHJpZ2dlclABgwUAJFEAgQBQFJlDb25maWdbMHgwMUZGMj'
    b'RdICYgMSA9PSAwQILlUAWVUHJlc3N1cmUgVHJpZ2dlciwgTG93UAGDAwAkUQGDAV+QUCWC'
    b'UGFRIYMB1MBQFbNTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaWRlIHByZXNzdX'
    b'JlIHdpbmRvdy5AguZQBZZQcmVzc3VyZSBUcmlnZ2VyLCBIaWdoUAGDBAAkUQGDAa2wUCWC'
    b'UGFRIYMB1MBQFbNTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaWRlIHByZXNzdX'
    b'JlIHdpbmRvdy5AGEEXUAWTVGVtcGVyYXR1cmUgVHJpZ2dlclABgwUBJFEAgQBQFJlDb25m'
    b'aWdbMHgwMUZGMjRdICYgMiA9PSAwQGPqUAWYVGVtcGVyYXR1cmUgVHJpZ2dlciwgTG93UA'
    b'GDAwEkUQKB9lAlgUNREoHYUSKBUFAVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91'
    b'dHNpZGUgdGVtcGVyYXR1cmUgd2luZG93LkBj61AFmVRlbXBlcmF0dXJlIFRyaWdnZXIsIE'
    b'hpZ2hQAYMEASRRAoEjUCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdo'
    b'ZW4gb3V0c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQBhBo1AFm0FjY2VsZXJhdGlvbiBUcm'
    b'lnZ2VyICgyMDBnKVABgwf/IFAV1ExvdyBwb3dlciB0cmlnZ2VyIHVzaW5nIGRpZ2l0YWwg'
    b'YWNjZWxlcm9tZXRlciB3aXRoIERDIHJlc3BvbnNlLiBObyBwcmUtdHJpZ2dlciBkYXRhLl'
    b'EAgQBQEYMF/whQEYMH/1BQJIBQFJRDb25maWdbMHgxZmYyMF0gPT0gMEBDQLFQBZZBY2Nl'
    b'bGVyYXRpb24gVGhyZXNob2xkUAGDBP8gUQGBUFIDiD/o9cKPXCj2USOIQGkAAAAAAABRE4'
    b'BQFfJMb3cgcG93ZXIgdHJpZ2dlci4gVGhlIG1pbmltdW0gYWNjZWxlcmF0aW9uIChwb3Np'
    b'dGl2ZSBvciBuZWdhdGl2ZSkgdG8gdHJpZ2dlciBhIHJlY29yZGluZy4gTXVzdCBiZSBncm'
    b'VhdGVyIHRoYW4gMC5AJ8ZQAYMF/yBRAYEHQQeRUAWOWCBBeGlzIFRyaWdnZXJBB5FQBY5Z'
    b'IEF4aXMgVHJpZ2dlckEHkVAFjlogQXhpcyBUcmlnZ2VyQBhBo1AFmkFjY2VsZXJhdGlvbi'
    b'BUcmlnZ2VyICg0MGcpUAGDB/9QUBXUTG93IHBvd2VyIHRyaWdnZXIgdXNpbmcgZGlnaXRh'
    b'bCBhY2NlbGVyb21ldGVyIHdpdGggREMgcmVzcG9uc2UuIE5vIHByZS10cmlnZ2VyIGRhdG'
    b'EuUQCBAFARgwX/CFARgwf/IFAkgFAUlENvbmZpZ1sweDFmZjUwXSA9PSAwQENAslAFlkFj'
    b'Y2VsZXJhdGlvbiBUaHJlc2hvbGRQAYME/1BRAYIgAFIDiD9EABQAFAAUUSOIQEQAAAAAAA'
    b'BRE4BQFfJMb3cgcG93ZXIgdHJpZ2dlci4gVGhlIG1pbmltdW0gYWNjZWxlcmF0aW9uIChw'
    b'b3NpdGl2ZSBvciBuZWdhdGl2ZSkgdG8gdHJpZ2dlciBhIHJlY29yZGluZy4gTXVzdCBiZS'
    b'BncmVhdGVyIHRoYW4gMC5AJ8ZQAYMF/1BRAYEHQQeRUAWOWCBBeGlzIFRyaWdnZXJBB5FQ'
    b'BY5ZIEF4aXMgVHJpZ2dlckEHkVAFjlogQXhpcyBUcmlnZ2VyQA+AQE+AQChGglAFjE1lYX'
    b'N1cmVtZW50c0AIQLtQBZwyMDBnIERDIEFjY2VsZXJhdGlvbiAoQ2ggMzIpQCefUAGDAf8g'
    b'UQGBAUEHklAFj0VuYWJsZSBBbGwgQXhlc0AX91AFi1NhbXBsZSBSYXRlUAGDAv8gUQGCAZ'
    b'BQJYJIelAUk25vdCBDb25maWdbMHgxZmYyMF1BB4VRAYIMgEEHhVEBggZAQQeFUQGCAyBB'
    b'B4VRAYIBkEEHhFEBgchBB4RRAYFkQQeEUQGBMkEHhFEBgRlBB4RRAYEMQAhBHFAFmzQwZy'
    b'BEQyBBY2NlbGVyYXRpb24gKENoIDgwKUAnn1ABgwH/UFEBgQFBB5JQBY9FbmFibGUgQWxs'
    b'IEF4ZXNAF0DYUAWLU2FtcGxlIFJhdGVQAYMC/1BRAYIB9FAlgkh6UBXeTG93IHBhc3MgZm'
    b'lsdGVyIGlzIHNldCB0byAyNSUgb2YgZGF0YSByYXRlLCBzbyBhIGRhdGEgcmF0ZSBvZiAy'
    b'MDAwSHogaGFzIGEgTFAgZmlsdGVyIGF0IDUwMCBIelAUk25vdCBDb25maWdbMHgxZmY1MF'
    b'1BB4VRAYIPoEEHhVEBggfQQQeFUQGCA+hBB4VRAYIB9EEHhFEBgfpBB4RRAYF9QQeEUQGB'
    b'P0EHhFEBgSBBB4RRAYEQQAhBMFAFqENvbnRyb2wgUGFkIFRlbXBlcmF0dXJlL1ByZXNzdX'
    b'JlIChDaCA1OSlAJ0CbUQGBA1ABgwH/O1AVpUVuYWJsZS9EaXNhYmxlIHNlbnNvcnMgb24g'
    b'Y29udHJvbCBwYWRBB5xQBZlFbmFibGUgUHJlc3N1cmUgKENoIDU5LjApQQefUAWcRW5hYm'
    b'xlIFRlbXBlcmF0dXJlIChDaCA1OS4xKUEHpVAFokVuYWJsZSBSZWxhdGl2ZSBIdW1pZGl0'
    b'eSAoQ2ggNTkuMilAEeNQBYtTYW1wbGUgUmF0ZVABgwL/O1AUlENvbmZpZ1sweDFmZjNCXS'
    b'A9PSAwUBWkU2V0IHNhbXBsZSBmcmVxdWVuY3kgZnJvbSAxIHRvIDEwIEh6UQGBClAlgkh6'
    b'URGBAVEhgQpACEHNUAWZSW5lcnRpYWwgTWVhc3VyZW1lbnQgVW5pdEAHQJZQBZBBY3F1aX'
    b'NpdGlvbiBNb2RlUAGDAf8rUCWBIFEBgRBBB4pQBYNPZmZRAYEAQQeoUAWhQWJzb2x1dGUg'
    b'T3JpZW50YXRpb24gKFF1YXRlcm5pb24pUQGBCEEHqFAFoVJlbGF0aXZlIE9yaWVudGF0aW'
    b'9uIChRdWF0ZXJuaW9uKVEBgRBBB49QBYhSb3RhdGlvblEBgQJAAUCGUAWVT3JpZW50YXRp'
    b'b24gRGF0YSBSYXRlUAGDAggrUQGBZFAlgkh6URGBAVEhgchQFb1TYW1wbGUgcmF0ZSBmb3'
    b'IgZGlyZWN0aW9uYWwgZGF0YSBmcm9tIDEgdG8gMjAwIEh6LiBDaGFubmVsIDY1UBSUQ29u'
    b'ZmlnWzB4MDFGRjJCXSA8IDhAAUCJUAWSUm90YXRpb24gRGF0YSBSYXRlUAGDAgIrUQGBZF'
    b'Algkh6URGBAVEhgchQFb5TZXQgc2FtcGxlIHJhdGUgZm9yIHJvdGF0aW9uIGRhdGEgZnJv'
    b'bSAxIHRvIDIwMCBIei4gQ2hhbm5lbCA0N1AUmUNvbmZpZ1sweDAxRkYyQl0gJiAyID09ID'
    b'BACEDxUAWlSW50ZXJuYWwgVGVtcGVyYXR1cmUvUHJlc3N1cmUgKENoIDM2KUAnQMVRAYED'
    b'UAGDAf8kUBX3RW5hYmxlL0Rpc2FibGUgc2xvd2VyIGludGVybmFsIGVudmlyb25tZW50YW'
    b'wgc2Vuc29ycy4gTk9URTogVGVtcGVyYXR1cmUgaXMgcmVxdWlyZWQgZm9yIG1haW4gYWNj'
    b'ZWxlcm9tZXRlciBtZWFzdXJlbWVudHNBB5xQBZlFbmFibGUgUHJlc3N1cmUgKENoIDM2Lj'
    b'ApQQefUAWcRW5hYmxlIFRlbXBlcmF0dXJlIChDaCAzNi4xKUAIQJBQBaBDb250cm9sIFBh'
    b'ZCBMaWdodCBTZW5zb3IgKENoIDc2KUAn6lABgwH/TFAVvkVuYWJsZS9EaXNhYmxlIGxpZ2'
    b'h0IHNlbnNvciBhdCA0IEh6IC0gTk9URTogRElTQUJMRUQgRk9SIEFMUEhBQQegUAWdRW5h'
    b'YmxlIExpZ2h0IFNlbnNvciAoQ2ggNzYuMClAD4BAT4BKCIBKKIBKSIA='
)



"""
Classes representing older Mide SlamStick data recorders.
"""

__author__ = "dstokes"
__copyright__ = "Copyright 2023 Mide Technology Corporation"

import os.path
import re
from typing import Union

from .base import Recorder
from .endaq import EndaqS
from .types import Filename

#===============================================================================
#
#===============================================================================


class SlamStickX(Recorder):
    """ A Slam Stick X data recorder from Mide Technology Corporation.
    """
    SN_FORMAT = "SSX%07d"

    # Match all Slam Sticks; this should be last in the list of types to find.
    _NAME_PATTERN = re.compile(r"^Slam Stick.*")

    _FW_UPDATE_FILE = os.path.join("SYSTEM", 'firmware.bin')

    manufacturer = "Midé Technology Corporation"
    homepage = "https://endaq.com/collections/endaq-shock-recorders-vibration-data-logger-sensors"


#===============================================================================
#
#===============================================================================

class SlamStickC(SlamStickX):
    """ A Slam Stick C data recorder from Mide Technology Corporation. Also
        sold as enDAQ Sx-D16.
    """
    SN_FORMAT = "SSC%07d"
    _NAME_PATTERN = re.compile(r"(^Slam Stick C.*)|(^S[234]-D16)")

    @property
    def serial(self) -> str:
        """ The recorder's manufacturer-issued serial number (as string). """
        # Hacky bit to provide Sx-D16 the enDAQ S serial number format.
        if self._sn is None:
            if self.productName.endswith('-D16'):
                self.SN_FORMAT = EndaqS.SN_FORMAT
        return Recorder.serial.fget(self)


    @property
    def path(self) -> Union[str, None]:
        return Recorder.path.fget(self)


    @path.setter
    def path(self, dev: Union[Filename, None]):
        """ The recorder's filesystem path (e.g., drive letter or mount point).
        """
        with self._busy:
            Recorder.path.fset(self, dev)
            if self._path:
                if self.getInfo('McuType', '').startswith("STM32"):
                    # HACK: New STM32-based Sx-D16 devices mostly emulate the
                    #  earlier EFM32 series 0 SlamStick C, but update with
                    #  the new firmware packages (i.e., `update.pkg`). This
                    #  is a convenient place to set it.
                    self._FW_UPDATE_FILE = Recorder._FW_UPDATE_FILE


# ===============================================================================
#
# ===============================================================================

class SlamStickS(SlamStickX):
    """ A Slam Stick S data recorder from Mide Technology Corporation.
    """
    SN_FORMAT = "SSS%07d"
    _NAME_PATTERN = re.compile("^Slam Stick S.*")



"""
Functions for detecting, identifying, and retrieving information about
data-logging devices.
"""

__author__ = "David Stokes"
__copyright__ = "Copyright 2024 Mide Technology Corporation"

import logging
import os
from pathlib import Path
import string
from threading import RLock
from typing import Dict, List, Optional, Union
from weakref import WeakValueDictionary

import ebmlite.core
from idelib.dataset import Dataset

from .base import Recorder, os_specific
from .command_interfaces import SerialCommandInterface
from .devinfo import SerialDeviceInfo
from .exceptions import *
from .endaq import EndaqS, EndaqW
from .response_codes import DeviceStatusCode
from .slamstick import SlamStickX, SlamStickC, SlamStickS
from .types import Drive, Filename, Epoch

logger = logging.getLogger('endaq.device')
# logger.setLevel(logging.DEBUG)

# ============================================================================
#
# ============================================================================

__version__ = "1.3.1"

__all__ = ('CommandError', 'ConfigError', 'ConfigVersionError',
           'DeviceError', 'DeviceTimeout', 'UnsupportedFeature',
           'deviceChanged', 'findDevice', 'fromRecording',
           'getDevices', 'getRecorder', 'isRecorder', 'onRecorder',
           'Recorder', 'EndaqS', 'EndaqW', 'SlamStickX', 'SlamStickC',
           'SlamStickS')

# ============================================================================
# EBML schema path modification
# ============================================================================

SCHEMA_PATH = "{endaq.device}/schemata"

# Ensure the `idelib` schemata are in the schema path (for idelib <= 3.2.4)
if "{idelib}/schemata" not in ebmlite.core.SCHEMA_PATH:
    ebmlite.core.SCHEMA_PATH.insert(0, "{idelib}/schemata")

# Add this package's schema to `ebmlite` schema search path, after
# `idelib`'s. This is a workaround for issue with legacy schema installed by
# earlier versions (can probably be removed after beta).
if SCHEMA_PATH not in ebmlite.core.SCHEMA_PATH:
    _idx = ebmlite.core.SCHEMA_PATH.index("{idelib}/schemata") + 1
    ebmlite.core.SCHEMA_PATH.insert(_idx, SCHEMA_PATH)

# ============================================================================
#
# ============================================================================

# Known classes or recorder. Checks are performed in the specified order, so
# put the ones with more general `isRecorder()` methods (i.e. superclasses)
# after the more specific ones. `SlamStickC` is first, since it is now sold
# as Sx-D16 but has the old SlamStick hardware, but the naming convention
# matches that of `EndaqS`. The base `Recorder` should be last.
RECORDER_TYPES = [SlamStickC, EndaqS, EndaqW, SlamStickS, SlamStickX, Recorder]

# Cache of previously seen recorders, to prevent redundant instantiations.
# Keyed by the hash of the recorders DEVINFO (or equivalent).
RECORDERS = {}

# Another cache of recorders, keyed by serial number. Used when discovering
# remote devices that don't immediately have DEVINFO accessible.
RECORDERS_BY_SN = WeakValueDictionary()

# Max number of cached recorders. Probably not needed, but just in case.
RECORDER_CACHE_SIZE = 100

# Lock to prevent contention (primarily with the recorder cache). Several
# classes have their own 'busy' locks as well.
_module_busy = RLock()


# ============================================================================
# Platform-specific stuff. 
# ============================================================================

def getRecorder(path: Filename,
                update: bool = False,
                strict: bool = True) -> Union[Recorder, None]:
    """ Get a specific recorder by its path.

        :param path: The filesystem path to the recorder's root directory.
        :param update: If `True`, update the path of known devices if they
            have changed (e.g., their drive letter or mount point changed
            after a device reset).
        :param strict: If `False`, only the directory structure within `path`
            is used to identify a recorder. If `True`, non-FAT file systems
            will be automatically rejected.
        :return: An instance of a :class:`~.endaq.device.Recorder` subclass,
            or `None` if the path is not a recorder.
    """
    global RECORDERS, RECORDERS_BY_SN

    dev = None

    with _module_busy:
        for rtype in RECORDER_TYPES:
            if rtype.isRecorder(path, strict=strict):
                # Get existing recorder if it has already been instantiated.
                devhash = rtype._getHash(path)
                dev = RECORDERS.pop(devhash, None)
                if not dev:
                    dev = rtype(path, strict=strict)
                else:
                    # Clear DEVINFO-getter, in case device was previously remote
                    dev._devinfo = None

                if devhash:
                    RECORDERS[devhash] = dev

                    # Path has changed. Note that the hash does not include
                    # path, in case a device rebooted and remounted with a
                    # different mount point/drive letter.
                    if update and dev.path != path:
                        dev.path = path

                RECORDERS_BY_SN[dev.serialInt] = dev

                break

        # Remove old cached devices. Ordered dictionaries assumed!
        if len(RECORDERS) > RECORDER_CACHE_SIZE:
            for k in list(RECORDERS.keys())[-RECORDER_CACHE_SIZE:]:
                del RECORDERS[k]

        return dev


def deviceChanged(recordersOnly: bool = True,
                  clear: bool = False) -> bool:
    """ Returns `True` if a drive has been connected or disconnected since
        the last call to :meth:`~.endaq.device.deviceChanged`.
        
        :param recordersOnly: If `False`, any change to the mounted drives
            is reported as a change. If `True`, the mounted drives are checked
            and `True` is only returned if the change occurred to a recorder.
            Checking for recorders only takes marginally more time.
        :param clear: If `True`, clear the cache of previously-detected
            drives and devices.
    """
    return os_specific.deviceChanged(recordersOnly, RECORDER_TYPES, clear=clear)


def getDeviceList(strict: bool = True) -> List[Drive]:
    """ Get a list of local data recorders, as their respective path (or the
        drive letter under Windows).

        :param strict: If `False`, only the directory structure is used
            to identify a recorder. If `True`, non-FAT file systems will
            be automatically rejected.
        :return: A list of `Drive` objects (named tuples containing the
            drive path, label, and other low-level filesystem info).
    """
    return os_specific.getDeviceList(RECORDER_TYPES, strict=strict)


def getDevices(paths: Optional[List[Filename]] = None,
               unmounted: bool = True,
               update: bool = True,
               strict: bool = True) -> List[Recorder]:
    """ Get a list of data recorder objects.
    
        :param paths: A list of specific paths to recording devices.
            Defaults to all found devices (as returned by
            :meth:`~.endaq.device.getDeviceList`).
        :param unmounted: If `True`, include devices connected by USB
            and responsive to commands but not appearing as drives.
            Note: Not all devices/firmware versions support this.
        :param update: If `True`, update the path of known devices if they
            have changed (e.g., their drive letter or mount point changed
            after a device reset).
        :param strict: If `False`, only the directory structure is used
            to identify a recorder. If `True`, non-FAT file systems and
            non-removable media will be automatically rejected.
        :return: A list of instances of `Recorder` subclasses.
    """
    global RECORDERS, RECORDERS_BY_SN

    with _module_busy:
        if paths is None:
            paths = getDeviceList(strict=strict)
        else:
            if isinstance(paths, (str, bytes, bytearray, Path)):
                paths = [paths]

        result = set()

        for path in paths:
            dev = getRecorder(path, update=update, strict=strict)
            if dev is not None:
                result.add(dev)

        if unmounted:
            for dev in getSerialDevices(known=RECORDERS_BY_SN):
                if not dev.available:
                    dev.path = None
                result.add(dev)
                RECORDERS.pop(hash(dev), None)
                RECORDERS[hash(dev)] = dev
                RECORDERS_BY_SN[dev.serialInt] = dev

        return sorted(result, key=lambda x: x.path or '\uffff')


def findDevice(sn: Optional[Union[str, int]] = None,
               chipId: Optional[Union[str, int]] = None,
               paths: Optional[List[Filename]] = None,
               unmounted: bool = False,
               update: bool = False,
               strict: bool = True) -> Union[Recorder, None]:
    """ Find a specific recorder by serial number or unique chip ID. One or
        the other must be provided, but not both. Note that early firmware
        versions do not report the device's chip ID.

        :param sn: The serial number of the recorder to find. Cannot be used
            with `chipId`. It can be an integer or a formatted serial number
            string (e.g., `12345` or `"S00012345"`).
        :param chipId: The chip ID of the recorder to find. Cannot be used
            with `sn`. It can be an integer or a hex string. Note that
            `chipId` cannot be used to find SlamStick and older enDAQ S
            devices (prior to hardware revision 2.0), as they do not report
            their chip ID.
        :param paths: A list of specific paths to recording devices.
            Defaults to all found devices (as returned by
            :meth:`~.endaq.device.getDeviceList`).
        :param unmounted: If `True`, include devices connected by USB
            and responsive to commands but not appearing as drives.
            Note: Not all devices/firmware versions support this.
        :param update: If `True`, update the path of known devices if they
            have changed (e.g., their drive letter or mount point changed
            after a device reset).
        :param strict: If `False`, only the directory structure is used
            to identify a recorder. If `True`, non-FAT file systems will
            be automatically rejected.
        :return: An instance of a :class:`~.endaq.device.Recorder` subclass
            representing the device with the specified serial number or chip
            ID, or `None` if it cannot be found.
    """
    with _module_busy:
        if sn and chipId:
            raise ValueError('Either a serial number or chip ID is required, not both')
        elif sn is None and chipId is None:
            raise ValueError('Either a serial number or chip ID is required')

        if isinstance(sn, str):
            sn = sn.lstrip(string.ascii_letters+"0")
            if not sn:
                sn = 0
            sn = int(sn)

        if isinstance(chipId, str):
            chipId = int(chipId, 16)

        for d in getDevices(paths, update=update, strict=strict, unmounted=unmounted):
            if sn is not None and d.serialInt == sn:
                return d
            elif chipId is not None and d.chipId == chipId:
                return d

        return None


# ============================================================================
# 
# ============================================================================

def isRecorder(path: Filename,
               strict: bool = True) -> bool:
    """ Determine if the given path is a recording device.

        :param path: The filesystem path to check.
        :param strict: If `False`, only the directory structure within `path`
            is used to identify a recorder. If `True`, non-FAT file systems
            will be automatically rejected.
    """
    with _module_busy:
        for t in RECORDER_TYPES:
            if t.isRecorder(path, strict=strict):
                return True
        return False


def onRecorder(path: Filename, strict: bool = True) -> bool:
    """ Returns the root directory of a recorder from a path to a directory or
        file on that recorder. It can be used to test whether a file is on
        a recorder. `False` is returned if the path is not on a recorder.
        The test is only whether the path refers to a recorder, not whether
        the path or file actually exists; if you need to know if the path
        is valid, perform your own checks first.
        
        :param path: The full path/name of a file.
        :param strict: If `False`, only the directory structure within `path`
            is used to identify a recorder. If `True`, non-FAT file systems
            will be automatically rejected.
        :return: The path to the root directory of a recorder (e.g. the drive
            letter in Windows) if the path is to a subdirectory on a recording 
            device, `False` if not.
    """
    oldp = None
    path = os.path.realpath(path)
    while path != oldp:
        if isRecorder(path, strict=strict):
            return path
        oldp = path
        path = os.path.dirname(path)
    return False


def fromRecording(doc: Dataset) -> Recorder:
    """ Create a 'virtual' recorder from the data contained in a recording
        file.

        :param doc: An imported IDE recording. Note that very old IDE files
            may not contain the metadata requires to create a `virtual`
            device.
    """
    productName = doc.recorderInfo.get('ProductName')
    if not productName:
        productName = doc.recorderInfo.get('PartNumber')
    if productName is None:
        raise TypeError("Could not create virtual recorder from file "
                        "(no ProductName or PartNumber in metadata)")
    recType = None
    for rec in RECORDER_TYPES:
        if rec._matchName(productName):
            recType = rec
            break
    if recType is None:
        return None
    return recType.fromRecording(doc)


# ============================================================================
# 
# ============================================================================

def getSerialDevices(known: Optional[Dict[int, Recorder]] = None,
                     strict: bool = True) -> List[Recorder]:
    """ Find all recorders with a serial command interface (and firmware
        that supports retrieving device metadata via that interface).

        :param known: A dictionary of known `Recorder` instances, keyed by
            device serial number.
        :param strict: If `True`, check the USB serial port VID and PID to
            see if they belong to a known type of device.
        :return: A list of `Recorder` instances found.
    """
    if known is None:
        known = {}

    devices = []

    # Dummy recorder and command interface to retrieve DEVINFO
    fake = Recorder(None)
    fake.command = SerialCommandInterface(fake)

    for port, sn in SerialCommandInterface._possibleRecorders(strict=strict):
        if sn in known:
            devices.append(known[sn])
            continue

        fake.command.port = None
        fake._snInt, fake._sn = sn, str(sn)

        with _module_busy:
            try:
                logger.debug(f'Getting info for SN {sn} via serial')
                info = fake.command._getInfo(0, index=False)
                if not info:
                    logger.debug(f'No info returned by SN {sn}, continuing')
                    continue
                for devtype in RECORDER_TYPES:
                    if devtype._isRecorder(info):
                        device = devtype(None, devinfo=info)
                        device.command = SerialCommandInterface(device)
                        device._devinfo = SerialDeviceInfo(device)
                        devices.append(device)
                        break
            except CommandError as err:
                if err.errno != DeviceStatusCode.ERR_INVALID_COMMAND:
                    logger.debug(f'Unexpected {type(err).__name__} getting info for {sn}: {err}')
                continue

    return devices
